# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from google.cloud.resourcemanager_v3 import gapic_version as package_version

__version__ = package_version.__version__


from .services.folders import FoldersAsyncClient, FoldersClient
from .services.organizations import OrganizationsAsyncClient, OrganizationsClient
from .services.projects import ProjectsAsyncClient, ProjectsClient
from .services.tag_bindings import TagBindingsAsyncClient, TagBindingsClient
from .services.tag_holds import TagHoldsAsyncClient, TagHoldsClient
from .services.tag_keys import TagKeysAsyncClient, TagKeysClient
from .services.tag_values import TagValuesAsyncClient, TagValuesClient
from .types.folders import (
    CreateFolderMetadata,
    CreateFolderRequest,
    DeleteFolderMetadata,
    DeleteFolderRequest,
    Folder,
    GetFolderRequest,
    ListFoldersRequest,
    ListFoldersResponse,
    MoveFolderMetadata,
    MoveFolderRequest,
    SearchFoldersRequest,
    SearchFoldersResponse,
    UndeleteFolderMetadata,
    UndeleteFolderRequest,
    UpdateFolderMetadata,
    UpdateFolderRequest,
)
from .types.organizations import (
    DeleteOrganizationMetadata,
    GetOrganizationRequest,
    Organization,
    SearchOrganizationsRequest,
    SearchOrganizationsResponse,
    UndeleteOrganizationMetadata,
)
from .types.projects import (
    CreateProjectMetadata,
    CreateProjectRequest,
    DeleteProjectMetadata,
    DeleteProjectRequest,
    GetProjectRequest,
    ListProjectsRequest,
    ListProjectsResponse,
    MoveProjectMetadata,
    MoveProjectRequest,
    Project,
    SearchProjectsRequest,
    SearchProjectsResponse,
    UndeleteProjectMetadata,
    UndeleteProjectRequest,
    UpdateProjectMetadata,
    UpdateProjectRequest,
)
from .types.tag_bindings import (
    CreateTagBindingMetadata,
    CreateTagBindingRequest,
    DeleteTagBindingMetadata,
    DeleteTagBindingRequest,
    EffectiveTag,
    ListEffectiveTagsRequest,
    ListEffectiveTagsResponse,
    ListTagBindingsRequest,
    ListTagBindingsResponse,
    TagBinding,
)
from .types.tag_holds import (
    CreateTagHoldMetadata,
    CreateTagHoldRequest,
    DeleteTagHoldMetadata,
    DeleteTagHoldRequest,
    ListTagHoldsRequest,
    ListTagHoldsResponse,
    TagHold,
)
from .types.tag_keys import (
    CreateTagKeyMetadata,
    CreateTagKeyRequest,
    DeleteTagKeyMetadata,
    DeleteTagKeyRequest,
    GetNamespacedTagKeyRequest,
    GetTagKeyRequest,
    ListTagKeysRequest,
    ListTagKeysResponse,
    Purpose,
    TagKey,
    UpdateTagKeyMetadata,
    UpdateTagKeyRequest,
)
from .types.tag_values import (
    CreateTagValueMetadata,
    CreateTagValueRequest,
    DeleteTagValueMetadata,
    DeleteTagValueRequest,
    GetNamespacedTagValueRequest,
    GetTagValueRequest,
    ListTagValuesRequest,
    ListTagValuesResponse,
    TagValue,
    UpdateTagValueMetadata,
    UpdateTagValueRequest,
)

__all__ = (
    "FoldersAsyncClient",
    "OrganizationsAsyncClient",
    "ProjectsAsyncClient",
    "TagBindingsAsyncClient",
    "TagHoldsAsyncClient",
    "TagKeysAsyncClient",
    "TagValuesAsyncClient",
    "CreateFolderMetadata",
    "CreateFolderRequest",
    "CreateProjectMetadata",
    "CreateProjectRequest",
    "CreateTagBindingMetadata",
    "CreateTagBindingRequest",
    "CreateTagHoldMetadata",
    "CreateTagHoldRequest",
    "CreateTagKeyMetadata",
    "CreateTagKeyRequest",
    "CreateTagValueMetadata",
    "CreateTagValueRequest",
    "DeleteFolderMetadata",
    "DeleteFolderRequest",
    "DeleteOrganizationMetadata",
    "DeleteProjectMetadata",
    "DeleteProjectRequest",
    "DeleteTagBindingMetadata",
    "DeleteTagBindingRequest",
    "DeleteTagHoldMetadata",
    "DeleteTagHoldRequest",
    "DeleteTagKeyMetadata",
    "DeleteTagKeyRequest",
    "DeleteTagValueMetadata",
    "DeleteTagValueRequest",
    "EffectiveTag",
    "Folder",
    "FoldersClient",
    "GetFolderRequest",
    "GetNamespacedTagKeyRequest",
    "GetNamespacedTagValueRequest",
    "GetOrganizationRequest",
    "GetProjectRequest",
    "GetTagKeyRequest",
    "GetTagValueRequest",
    "ListEffectiveTagsRequest",
    "ListEffectiveTagsResponse",
    "ListFoldersRequest",
    "ListFoldersResponse",
    "ListProjectsRequest",
    "ListProjectsResponse",
    "ListTagBindingsRequest",
    "ListTagBindingsResponse",
    "ListTagHoldsRequest",
    "ListTagHoldsResponse",
    "ListTagKeysRequest",
    "ListTagKeysResponse",
    "ListTagValuesRequest",
    "ListTagValuesResponse",
    "MoveFolderMetadata",
    "MoveFolderRequest",
    "MoveProjectMetadata",
    "MoveProjectRequest",
    "Organization",
    "OrganizationsClient",
    "Project",
    "ProjectsClient",
    "Purpose",
    "SearchFoldersRequest",
    "SearchFoldersResponse",
    "SearchOrganizationsRequest",
    "SearchOrganizationsResponse",
    "SearchProjectsRequest",
    "SearchProjectsResponse",
    "TagBinding",
    "TagBindingsClient",
    "TagHold",
    "TagHoldsClient",
    "TagKey",
    "TagKeysClient",
    "TagValue",
    "TagValuesClient",
    "UndeleteFolderMetadata",
    "UndeleteFolderRequest",
    "UndeleteOrganizationMetadata",
    "UndeleteProjectMetadata",
    "UndeleteProjectRequest",
    "UpdateFolderMetadata",
    "UpdateFolderRequest",
    "UpdateProjectMetadata",
    "UpdateProjectRequest",
    "UpdateTagKeyMetadata",
    "UpdateTagKeyRequest",
    "UpdateTagValueMetadata",
    "UpdateTagValueRequest",
)



# Copyright 2020 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-API Connection for the Google Cloud Spanner."""
import warnings

from google.api_core.client_options import ClientOptions
from google.api_core.exceptions import Aborted
from google.api_core.gapic_v1.client_info import ClientInfo
from google.auth.credentials import AnonymousCredentials

from google.cloud import spanner_v1 as spanner
from google.cloud.spanner_dbapi import partition_helper
from google.cloud.spanner_dbapi.batch_dml_executor import BatchMode, BatchDmlExecutor
from google.cloud.spanner_dbapi.parsed_statement import AutocommitDmlMode
from google.cloud.spanner_dbapi.partition_helper import PartitionId
from google.cloud.spanner_dbapi.parsed_statement import ParsedStatement, Statement
from google.cloud.spanner_dbapi.transaction_helper import TransactionRetryHelper
from google.cloud.spanner_dbapi.cursor import Cursor
from google.cloud.spanner_v1 import RequestOptions, TransactionOptions
from google.cloud.spanner_v1.database_sessions_manager import TransactionType
from google.cloud.spanner_v1.snapshot import Snapshot

from google.cloud.spanner_dbapi.exceptions import (
    InterfaceError,
    OperationalError,
    ProgrammingError,
)
from google.cloud.spanner_dbapi.version import DEFAULT_USER_AGENT
from google.cloud.spanner_dbapi.version import PY_VERSION


CLIENT_TRANSACTION_NOT_STARTED_WARNING = (
    "This method is non-operational as a transaction has not been started."
)


def check_not_closed(function):
    """`Connection` class methods decorator.

    Raise an exception if the connection is closed.

    :raises: :class:`InterfaceError` if the connection is closed.
    """

    def wrapper(connection, *args, **kwargs):
        if connection.is_closed:
            raise InterfaceError("Connection is already closed")

        return function(connection, *args, **kwargs)

    return wrapper


class Connection:
    """Representation of a DB-API connection to a Cloud Spanner database.

    You most likely don't need to instantiate `Connection` objects
    directly, use the `connect` module function instead.

    :type instance: :class:`~google.cloud.spanner_v1.instance.Instance`
    :param instance: Cloud Spanner instance to connect to.

    :type database: :class:`~google.cloud.spanner_v1.database.Database`
    :param database: The database to which the connection is linked.

    :type read_only: bool
    :param read_only:
        Flag to indicate that the connection may only execute queries and no update or DDL statements.
        If True, the connection will use a single use read-only transaction with strong timestamp
        bound for each new statement, and will immediately see any changes that have been committed by
        any other transaction.
        If autocommit is false, the connection will automatically start a new multi use read-only transaction
        with strong timestamp bound when the first statement is executed. This read-only transaction will be
        used for all subsequent statements until either commit() or rollback() is called on the connection. The
        read-only transaction will read from a consistent snapshot of the database at the time that the
        transaction started. This means that the transaction will not see any changes that have been
        committed by other transactions since the start of the read-only transaction. Commit or rolling back
        the read-only transaction is semantically the same, and only indicates that the read-only transaction
        should end a that a new one should be started when the next statement is executed.

    **kwargs: Initial value for connection variables.
    """

    def __init__(self, instance, database=None, read_only=False, **kwargs):
        self._instance = instance
        self._database = database
        self._ddl_statements = []

        self._transaction = None
        self._session = None
        self._snapshot = None

        self.is_closed = False
        self._autocommit = False
        # indicator to know if the session pool used by
        # this connection should be cleared on the
        # connection close
        self._own_pool = True
        self._read_only = read_only
        self._staleness = None
        self.request_priority = None
        self._transaction_begin_marked = False
        self._transaction_isolation_level = None
        # whether transaction started at Spanner. This means that we had
        # made at least one call to Spanner.
        self._spanner_transaction_started = False
        self._batch_mode = BatchMode.NONE
        self._batch_dml_executor: BatchDmlExecutor = None
        self._transaction_helper = TransactionRetryHelper(self)
        self._autocommit_dml_mode: AutocommitDmlMode = AutocommitDmlMode.TRANSACTIONAL
        self._connection_variables = kwargs

    @property
    def spanner_client(self):
        """Client for interacting with Cloud Spanner API. This property exposes
        the spanner client so that underlying methods can be accessed.
        """
        return self._instance._client

    @property
    def current_schema(self):
        """schema name for this connection.

        :rtype: str
        :returns: the current default schema of this connection. Currently, this
         is always "" for GoogleSQL and "public" for PostgreSQL databases.
        """
        if self.database is None:
            raise ValueError("database property not set on the connection")
        return self.database.default_schema_name

    @property
    def autocommit(self):
        """Autocommit mode flag for this connection.

        :rtype: bool
        :returns: Autocommit mode flag value.
        """
        return self._autocommit

    @autocommit.setter
    def autocommit(self, value):
        """Change this connection autocommit mode. Setting this value to True
        while a transaction is active will commit the current transaction.

        :type value: bool
        :param value: New autocommit mode state.
        """
        if value and not self._autocommit and self._spanner_transaction_started:
            self.commit()

        self._autocommit = value

    @property
    def database(self):
        """Database to which this connection relates.

        :rtype: :class:`~google.cloud.spanner_v1.database.Database`
        :returns: The related database object.
        """
        return self._database

    @property
    def autocommit_dml_mode(self):
        """Modes for executing DML statements in autocommit mode for this connection.

        The DML autocommit modes are:
        1) TRANSACTIONAL - DML statements are executed as single read-write transaction.
        After successful execution, the DML statement is guaranteed to have been applied
        exactly once to the database.

        2) PARTITIONED_NON_ATOMIC - DML statements are executed as partitioned DML transactions.
        If an error occurs during the execution of the DML statement, it is possible that the
        statement has been applied to some but not all of the rows specified in the statement.

        :rtype: :class:`~google.cloud.spanner_dbapi.parsed_statement.AutocommitDmlMode`
        """
        return self._autocommit_dml_mode

    @property
    def inside_transaction(self):
        warnings.warn(
            "This method is deprecated. Use _spanner_transaction_started field",
            DeprecationWarning,
        )
        return (
            self._transaction
            and not self._transaction.committed
            and not self._transaction.rolled_back
        )

    @property
    def _client_transaction_started(self):
        """Flag: whether transaction started at client side.

        Returns:
            bool: True if transaction started, False otherwise.
        """
        return (not self._autocommit) or self._transaction_begin_marked

    @property
    def _ignore_transaction_warnings(self):
        return self._connection_variables.get("ignore_transaction_warnings", False)

    @property
    def instance(self):
        """Instance to which this connection relates.

        :rtype: :class:`~google.cloud.spanner_v1.instance.Instance`
        :returns: The related instance object.
        """
        return self._instance

    @property
    def read_only(self):
        """Flag: the connection can be used only for database reads.

        Returns:
            bool:
                True if the connection may only be used for database reads.
        """
        return self._read_only

    @read_only.setter
    def read_only(self, value):
        """`read_only` flag setter.

        Args:
            value (bool): True for ReadOnly mode, False for ReadWrite.
        """
        if self._read_only != value and self._spanner_transaction_started:
            raise ValueError(
                "Connection read/write mode can't be changed while a transaction is in progress. "
                "Commit or rollback the current transaction and try again."
            )
        self._read_only = value

    @property
    def request_options(self):
        """Options for the next SQL operations.

        Returns:
            google.cloud.spanner_v1.RequestOptions:
                Request options.
        """
        if self.request_priority is None:
            return

        req_opts = RequestOptions(priority=self.request_priority)
        self.request_priority = None
        return req_opts

    @property
    def transaction_tag(self):
        """The transaction tag that will be applied to the next read/write
        transaction on this `Connection`. This property is automatically cleared
        when a new transaction is started.

        Returns:
            str: The transaction tag that will be applied to the next read/write transaction.
        """
        return self._connection_variables.get("transaction_tag", None)

    @transaction_tag.setter
    def transaction_tag(self, value):
        """Sets the transaction tag for the next read/write transaction on this
        `Connection`. This property is automatically cleared when a new transaction
        is started.

        Args:
            value (str): The transaction tag for the next read/write transaction.
        """
        self._connection_variables["transaction_tag"] = value

    @property
    def isolation_level(self):
        """The default isolation level that is used for all read/write
        transactions on this `Connection`.

        Returns:
            google.cloud.spanner_v1.types.TransactionOptions.IsolationLevel:
            The isolation level that is used for read/write transactions on
            this `Connection`.
        """
        return self._connection_variables.get(
            "isolation_level",
            TransactionOptions.IsolationLevel.ISOLATION_LEVEL_UNSPECIFIED,
        )

    @isolation_level.setter
    def isolation_level(self, value: TransactionOptions.IsolationLevel):
        """Sets the isolation level that is used for all read/write
        transactions on this `Connection`.

        Args:
            value (google.cloud.spanner_v1.types.TransactionOptions.IsolationLevel):
            The isolation level for all read/write transactions on this
            `Connection`.
        """
        self._connection_variables["isolation_level"] = value

    @property
    def staleness(self):
        """Current read staleness option value of this `Connection`.

        Returns:
            dict: Staleness type and value.
        """
        return self._staleness or {}

    @staleness.setter
    def staleness(self, value):
        """Read staleness option setter.

        Args:
            value (dict): Staleness type and value.
        """
        if self._spanner_transaction_started and value != self._staleness:
            raise ValueError(
                "`staleness` option can't be changed while a transaction is in progress. "
                "Commit or rollback the current transaction and try again."
            )

        possible_opts = (
            "read_timestamp",
            "min_read_timestamp",
            "max_staleness",
            "exact_staleness",
        )
        if value is not None and sum([opt in value for opt in possible_opts]) != 1:
            raise ValueError(
                "Expected one of the following staleness options: "
                "read_timestamp, min_read_timestamp, max_staleness, exact_staleness."
            )

        self._staleness = value

    def _session_checkout(self):
        """Get a Cloud Spanner session from the pool.

        If there is already a session associated with
        this connection, it'll be used instead.

        :rtype: :class:`google.cloud.spanner_v1.session.Session`
        :returns: Cloud Spanner session object ready to use.
        """
        if self.database is None:
            raise ValueError("Database needs to be passed for this operation")

        if not self._session:
            transaction_type = (
                TransactionType.READ_ONLY
                if self.read_only
                else TransactionType.READ_WRITE
            )
            self._session = self.database._sessions_manager.get_session(
                transaction_type
            )

        return self._session

    def _release_session(self):
        """Release the currently used Spanner session.

        The session will be returned into the sessions pool.
        """
        if self._session is None:
            return

        if self.database is None:
            raise ValueError("Database needs to be passed for this operation")

        self.database._sessions_manager.put_session(self._session)
        self._session = None

    def transaction_checkout(self):
        """Get a Cloud Spanner transaction.

        Begin a new transaction, if there is no transaction in
        this connection yet. Return the started one otherwise.

        This method is a no-op if the connection is in autocommit mode and no
        explicit transaction has been started

        :rtype: :class:`google.cloud.spanner_v1.transaction.Transaction`
        :returns: A Cloud Spanner transaction object, ready to use.
        """
        if not self.read_only and self._client_transaction_started:
            if not self._spanner_transaction_started:
                self._transaction = self._session_checkout().transaction()
                self._transaction.transaction_tag = self.transaction_tag
                if self._transaction_isolation_level:
                    self._transaction.isolation_level = (
                        self._transaction_isolation_level
                    )
                else:
                    self._transaction.isolation_level = self.isolation_level
                self.transaction_tag = None
                self._snapshot = None
                self._spanner_transaction_started = True
                self._transaction.begin()

            return self._transaction

    def snapshot_checkout(self):
        """Get a Cloud Spanner snapshot.

        Initiate a new multi-use snapshot, if there is no snapshot in
        this connection yet. Return the existing one otherwise.

        :rtype: :class:`google.cloud.spanner_v1.snapshot.Snapshot`
        :returns: A Cloud Spanner snapshot object, ready to use.
        """
        if self.read_only and self._client_transaction_started:
            if not self._spanner_transaction_started:
                self._snapshot = Snapshot(
                    self._session_checkout(), multi_use=True, **self.staleness
                )
                self._transaction = None
                self._snapshot.begin()
                self._spanner_transaction_started = True

            return self._snapshot

    def close(self):
        """Closes this connection.

        The connection will be unusable from this point forward. If the
        connection has an active transaction, it will be rolled back.
        """
        if self._spanner_transaction_started and not self._read_only:
            self._transaction.rollback()

        if self._own_pool and self.database:
            self.database._sessions_manager._pool.clear()

        self.is_closed = True

    @check_not_closed
    def begin(self, isolation_level=None):
        """
        Marks the transaction as started.

        :raises: :class:`InterfaceError`: if this connection is closed.
        :raises: :class:`OperationalError`: if there is an existing transaction
        that has been started
        """
        if self._transaction_begin_marked:
            raise OperationalError("A transaction has already started")
        if self._spanner_transaction_started:
            raise OperationalError(
                "Beginning a new transaction is not allowed when a transaction "
                "is already running"
            )
        self._transaction_begin_marked = True
        self._transaction_isolation_level = isolation_level

    def commit(self):
        """Commits any pending transaction to the database.
        This is a no-op if there is no active client transaction.
        """
        if self.database is None:
            raise ValueError("Database needs to be passed for this operation")
        if not self._client_transaction_started:
            if not self._ignore_transaction_warnings:
                warnings.warn(
                    CLIENT_TRANSACTION_NOT_STARTED_WARNING, UserWarning, stacklevel=2
                )
            return

        self.run_prior_DDL_statements()
        try:
            if self._spanner_transaction_started and not self._read_only:
                self._transaction.commit()
        except Aborted:
            self._transaction_helper.retry_transaction()
            self.commit()
        finally:
            self._reset_post_commit_or_rollback()

    def rollback(self):
        """Rolls back any pending transaction.
        This is a no-op if there is no active client transaction.
        """
        if not self._client_transaction_started:
            if not self._ignore_transaction_warnings:
                warnings.warn(
                    CLIENT_TRANSACTION_NOT_STARTED_WARNING, UserWarning, stacklevel=2
                )
            return
        try:
            if self._spanner_transaction_started and not self._read_only:
                self._transaction.rollback()
        finally:
            self._reset_post_commit_or_rollback()

    def _reset_post_commit_or_rollback(self):
        self._release_session()
        self._transaction_helper.reset()
        self._transaction_begin_marked = False
        self._transaction_isolation_level = None
        self._spanner_transaction_started = False

    @check_not_closed
    def cursor(self):
        """Factory to create a DB API Cursor."""
        return Cursor(self)

    @check_not_closed
    def run_prior_DDL_statements(self):
        if self.database is None:
            raise ValueError("Database needs to be passed for this operation")
        if self._ddl_statements:
            ddl_statements = self._ddl_statements
            self._ddl_statements = []

            return self.database.update_ddl(ddl_statements).result()

    def run_statement(
        self, statement: Statement, request_options: RequestOptions = None
    ):
        """Run single SQL statement in begun transaction.

        This method is never used in autocommit mode. In
        !autocommit mode however it remembers every executed
        SQL statement with its parameters.

        :type statement: :class:`Statement`
        :param statement: SQL statement to execute.

        :type retried: bool
        :param retried: (Optional) Retry the SQL statement if statement
                        execution failed. Defaults to false.

        :type request_options: :class:`RequestOptions`
        :param request_options: Request options to use for this statement.

        :rtype: :class:`google.cloud.spanner_v1.streamed.StreamedResultSet`,
                :class:`google.cloud.spanner_dbapi.checksum.ResultsChecksum`
        :returns: Streamed result set of the statement and a
                  checksum of this statement results.
        """
        transaction = self.transaction_checkout()
        return transaction.execute_sql(
            statement.sql,
            statement.params,
            param_types=statement.param_types,
            request_options=request_options or self.request_options,
        )

    @check_not_closed
    def validate(self):
        """
        Execute a minimal request to check if the connection
        is valid and the related database is reachable.

        Raise an exception in case if the connection is closed,
        invalid, target database is not found, or the request result
        is incorrect.

        :raises: :class:`InterfaceError`: if this connection is closed.
        :raises: :class:`OperationalError`: if the request result is incorrect.
        :raises: :class:`google.cloud.exceptions.NotFound`: if the linked instance
                  or database doesn't exist.
        """
        if self.database is None:
            raise ValueError("Database needs to be passed for this operation")
        with self.database.snapshot() as snapshot:
            result = list(snapshot.execute_sql("SELECT 1"))
            if result != [[1]]:
                raise OperationalError(
                    "The checking query (SELECT 1) returned an unexpected result: %s. "
                    "Expected: [[1]]" % result
                )

    @check_not_closed
    def start_batch_dml(self, cursor):
        if self._batch_mode is not BatchMode.NONE:
            raise ProgrammingError(
                "Cannot start a DML batch when a batch is already active"
            )
        if self.read_only:
            raise ProgrammingError(
                "Cannot start a DML batch when the connection is in read-only mode"
            )
        self._batch_mode = BatchMode.DML
        self._batch_dml_executor = BatchDmlExecutor(cursor)

    @check_not_closed
    def execute_batch_dml_statement(self, parsed_statement: ParsedStatement):
        if self._batch_mode is not BatchMode.DML:
            raise ProgrammingError(
                "Cannot execute statement when the BatchMode is not DML"
            )
        self._batch_dml_executor.execute_statement(parsed_statement)

    @check_not_closed
    def run_batch(self):
        if self._batch_mode is BatchMode.NONE:
            raise ProgrammingError("Cannot run a batch when the BatchMode is not set")
        try:
            if self._batch_mode is BatchMode.DML:
                many_result_set = self._batch_dml_executor.run_batch_dml()
        finally:
            self._batch_mode = BatchMode.NONE
            self._batch_dml_executor = None
        return many_result_set

    @check_not_closed
    def abort_batch(self):
        if self._batch_mode is BatchMode.NONE:
            raise ProgrammingError("Cannot abort a batch when the BatchMode is not set")
        if self._batch_mode is BatchMode.DML:
            self._batch_dml_executor = None
        self._batch_mode = BatchMode.NONE

    @check_not_closed
    def partition_query(
        self,
        parsed_statement: ParsedStatement,
        query_options=None,
    ):
        statement = parsed_statement.statement
        partitioned_query = parsed_statement.client_side_statement_params[0]
        self._partitioned_query_validation(partitioned_query, statement)

        batch_snapshot = self._database.batch_snapshot()
        partition_ids = []
        partitions = list(
            batch_snapshot.generate_query_batches(
                partitioned_query,
                statement.params,
                statement.param_types,
                query_options=query_options,
            )
        )

        batch_transaction_id = batch_snapshot.get_batch_transaction_id()
        for partition in partitions:
            partition_ids.append(
                partition_helper.encode_to_string(batch_transaction_id, partition)
            )
        return partition_ids

    @check_not_closed
    def run_partition(self, encoded_partition_id):
        partition_id: PartitionId = partition_helper.decode_from_string(
            encoded_partition_id
        )
        batch_transaction_id = partition_id.batch_transaction_id
        batch_snapshot = self._database.batch_snapshot(
            read_timestamp=batch_transaction_id.read_timestamp,
            session_id=batch_transaction_id.session_id,
            transaction_id=batch_transaction_id.transaction_id,
        )
        return batch_snapshot.process(partition_id.partition_result)

    @check_not_closed
    def run_partitioned_query(
        self,
        parsed_statement: ParsedStatement,
    ):
        statement = parsed_statement.statement
        partitioned_query = parsed_statement.client_side_statement_params[0]
        self._partitioned_query_validation(partitioned_query, statement)
        batch_snapshot = self._database.batch_snapshot()
        return batch_snapshot.run_partitioned_query(
            partitioned_query, statement.params, statement.param_types
        )

    @check_not_closed
    def _set_autocommit_dml_mode(
        self,
        parsed_statement: ParsedStatement,
    ):
        autocommit_dml_mode_str = parsed_statement.client_side_statement_params[0]
        autocommit_dml_mode = AutocommitDmlMode[autocommit_dml_mode_str.upper()]
        self.set_autocommit_dml_mode(autocommit_dml_mode)

    def set_autocommit_dml_mode(
        self,
        autocommit_dml_mode,
    ):
        """
        Sets the mode for executing DML statements in autocommit mode for this connection.
        This mode is only used when the connection is in autocommit mode, and may only
        be set while the transaction is in autocommit mode and not in a temporary transaction.
        """

        if self._client_transaction_started is True:
            raise ProgrammingError(
                "Cannot set autocommit DML mode while not in autocommit mode or while a transaction is active."
            )
        if self.read_only is True:
            raise ProgrammingError(
                "Cannot set autocommit DML mode for a read-only connection."
            )
        if self._batch_mode is not BatchMode.NONE:
            raise ProgrammingError("Cannot set autocommit DML mode while in a batch.")
        self._autocommit_dml_mode = autocommit_dml_mode

    def _partitioned_query_validation(self, partitioned_query, statement):
        if self.read_only is not True and self._client_transaction_started is True:
            raise ProgrammingError(
                "Partitioned query is not supported, because the connection is in a read/write transaction."
            )

    def __enter__(self):
        return self

    def __exit__(self, etype, value, traceback):
        self.commit()
        self.close()


def connect(
    instance_id,
    database_id=None,
    project=None,
    credentials=None,
    pool=None,
    user_agent=None,
    client=None,
    route_to_leader_enabled=True,
    database_role=None,
    experimental_host=None,
    **kwargs,
):
    """Creates a connection to a Google Cloud Spanner database.

    :type instance_id: str
    :param instance_id: The ID of the instance to connect to.

    :type database_id: str
    :param database_id: (Optional) The ID of the database to connect to.

    :type project: str
    :param project: (Optional) The ID of the project which owns the
                    instances, tables and data. If not provided, will
                    attempt to determine from the environment.

    :type credentials: Union[:class:`~google.auth.credentials.Credentials`, str]
    :param credentials: (Optional) The authorization credentials to attach to
                        requests. These credentials identify this application
                        to the service. These credentials may be specified as
                        a file path indicating where to retrieve the service
                        account JSON for the credentials to connect to
                        Cloud Spanner. If none are specified, the client will
                        attempt to ascertain the credentials from the
                        environment.

    :type pool: Concrete subclass of
                :class:`~google.cloud.spanner_v1.pool.AbstractSessionPool`.
    :param pool: (Optional). Session pool to be used by database.

    :type user_agent: str
    :param user_agent: (Optional) User agent to be used with this connection's
                       requests.

    :type client: Concrete subclass of
                  :class:`~google.cloud.spanner_v1.Client`.
    :param client: (Optional) Custom user provided Client Object

    :type route_to_leader_enabled: boolean
    :param route_to_leader_enabled:
        (Optional) Default True. Set route_to_leader_enabled as False to
        disable leader aware routing. Disabling leader aware routing would
        route all requests in RW/PDML transactions to the closest region.

    :type database_role: str
    :param database_role: (Optional) The database role to connect as when using
        fine-grained access controls.

    **kwargs: Initial value for connection variables.


    :rtype: :class:`google.cloud.spanner_dbapi.connection.Connection`
    :returns: Connection object associated with the given Google Cloud Spanner
              resource.
    """
    if client is None:
        client_info = ClientInfo(
            user_agent=user_agent or DEFAULT_USER_AGENT,
            python_version=PY_VERSION,
            client_library_version=spanner.__version__,
        )
        if isinstance(credentials, str):
            client = spanner.Client.from_service_account_json(
                credentials,
                project=project,
                client_info=client_info,
                route_to_leader_enabled=route_to_leader_enabled,
            )
        else:
            client_options = None
            if isinstance(credentials, AnonymousCredentials):
                client_options = kwargs.get("client_options")
            if experimental_host is not None:
                project = "default"
                credentials = AnonymousCredentials()
                client_options = ClientOptions(api_endpoint=experimental_host)
            client = spanner.Client(
                project=project,
                credentials=credentials,
                client_info=client_info,
                route_to_leader_enabled=route_to_leader_enabled,
                client_options=client_options,
            )
    else:
        if project is not None and client.project != project:
            raise ValueError("project in url does not match client object project")

    instance = client.instance(instance_id)
    database = None
    if database_id:
        logger = kwargs.get("logger")
        database = instance.database(
            database_id, pool=pool, database_role=database_role, logger=logger
        )
    conn = Connection(instance, database, **kwargs)
    if pool is not None:
        conn._own_pool = False

    return conn



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from typing import Any, Callable, Dict, List, Mapping, Optional, TYPE_CHECKING


if TYPE_CHECKING:
    try:
        from a2a.server.request_handlers import RequestHandler
        from a2a.server.tasks import TaskStore
        from a2a.types import AgentCard, AgentSkill
        from a2a.server.agent_execution import AgentExecutor
        from a2a.server.context import ServerCallContext

        RequestHandler = RequestHandler
        TaskStore = TaskStore
        AgentCard = AgentCard
        AgentSkill = AgentSkill
        AgentExecutor = AgentExecutor
        ServerCallContext = ServerCallContext
    except (ImportError, AttributeError):
        RequestHandler = Any
        TaskStore = Any
        AgentCard = Any
        AgentSkill = Any
        AgentExecutor = Any
        ServerCallContext = Any

    try:
        from fastapi import Request

        Request = Request
    except (ImportError, AttributeError):
        Request = Any


def create_agent_card(
    agent_name: Optional[str] = None,
    description: Optional[str] = None,
    skills: Optional[List["AgentSkill"]] = None,
    agent_card: Optional[Dict[str, Any]] = None,
    default_input_modes: Optional[List[str]] = None,
    default_output_modes: Optional[List[str]] = None,
) -> "AgentCard":
    """Creates an AgentCard object.

    The function can be called in two ways:
    1. By providing the individual parameters: agent_name, description, and skills.
    2. By providing a single dictionary containing all the data.

    If a dictionary is provided, the other parameters are ignored.

    Args:
        agent_name (Optional[str]): The name of the agent.
        description (Optional[str]): A description of the agent.
        skills (Optional[List[AgentSkill]]): A list of AgentSkills.
        agent_card (Optional[Dict[str, Any]]): Agent Card as a dictionary.
        default_input_modes (Optional[List[str]]): A list of input modes,
            default to ["text/plain"].
        default_output_modes (Optional[List[str]]): A list of output modes,
            default to ["application/json"].

    Returns:
        AgentCard: A fully constructed AgentCard object.

    Raises:
        ValueError: If neither a dictionary nor the required parameters are provided.
    """
    # pylint: disable=g-import-not-at-top
    from a2a.types import AgentCard, AgentCapabilities, TransportProtocol

    # Check if a dictionary was provided.
    if agent_card:
        return AgentCard(**agent_card)

    # If no dictionary, use the individual parameters.
    elif agent_name and description and skills:
        return AgentCard(
            name=agent_name,
            description=description,
            url="http://localhost:9999/",
            version="1.0.0",
            default_input_modes=default_input_modes or ["text/plain"],
            default_output_modes=default_output_modes or ["application/json"],
            # Agent Engine does not support streaming yet
            capabilities=AgentCapabilities(streaming=False),
            skills=skills,
            preferred_transport=TransportProtocol.http_json,  # Http Only.
            supports_authenticated_extended_card=True,
        )

    # Raise an error if insufficient data is provided.
    else:
        raise ValueError(
            "Please provide either an agent_card or all of the required "
            "parameters (agent_name, description, and skills)."
        )


def default_a2a_agent() -> "A2aAgent":
    """Creates a default A2aAgent instance."""
    # pylint: disable=g-import-not-at-top
    from a2a.server.agent_execution import AgentExecutor, RequestContext
    from a2a.types import AgentSkill
    from a2a.server.events import EventQueue
    from a2a.utils import new_agent_text_message

    skill = AgentSkill(
        id="hello_world",
        name="Returns hello world",
        description="just returns hello world",
        tags=["hello world"],
        examples=["hi", "hello world"],
    )
    agent_card = create_agent_card(
        agent_name="Hello World Agent",
        description="Just a hello world agent",
        skills=[skill],
    )

    class HelloWorldAgentExecutor(AgentExecutor):
        """Hello World Agent Executor."""

        def get_agent_response(self) -> str:
            return "Hello World"

        async def execute(
            self,
            context: RequestContext,
            event_queue: EventQueue,
        ) -> None:
            result = self.get_agent_response()
            await event_queue.enqueue_event(new_agent_text_message(result))

        async def cancel(
            self, context: RequestContext, event_queue: EventQueue
        ) -> None:
            raise Exception("cancel not supported")

    return A2aAgent(
        agent_card=agent_card,
        agent_executor_builder=HelloWorldAgentExecutor,
    )


class A2aAgent:
    """A class to initialize and set up an Agent-to-Agent application."""

    # TODO: Add instrumentation for the A2A agent.
    def __init__(
        self,
        *,
        agent_card: "AgentCard",
        task_store_builder: Callable[..., "TaskStore"] = None,
        task_store_kwargs: Optional[Mapping[str, Any]] = None,
        agent_executor_kwargs: Optional[Mapping[str, Any]] = None,
        agent_executor_builder: Optional[Callable[..., "AgentExecutor"]] = None,
        request_handler_kwargs: Optional[Mapping[str, Any]] = None,
        request_handler_builder: Optional[Callable[..., "RequestHandler"]] = None,
        extended_agent_card: "AgentCard" = None,
    ):
        """Initializes the A2A agent."""
        # pylint: disable=g-import-not-at-top
        from google.cloud.aiplatform import initializer
        from a2a.types import TransportProtocol

        if (
            agent_card.preferred_transport
            and agent_card.preferred_transport != TransportProtocol.http_json
        ):
            raise ValueError(
                "Only HTTP+JSON is supported for preferred transport on agent card "
            )
        if agent_card.capabilities and agent_card.capabilities.streaming:
            raise ValueError("Streaming is not supported by Agent Engine")

        self._tmpl_attrs: dict[str, Any] = {
            "project": initializer.global_config.project,
            "location": initializer.global_config.location,
            "agent_card": agent_card,
            "agent_executor": None,
            "agent_executor_kwargs": agent_executor_kwargs or {},
            "agent_executor_builder": agent_executor_builder,
            "task_store": None,
            "task_store_kwargs": task_store_kwargs or {},
            "task_store_builder": task_store_builder,
            "request_handler": None,
            "request_handler_kwargs": request_handler_kwargs or {},
            "request_handler_builder": request_handler_builder,
            "extended_agent_card": extended_agent_card,
        }
        self.agent_card = agent_card
        self.a2a_rest_adapter = None
        self.request_handler = None
        self.rest_handler = None
        self.task_store = None
        self.agent_executor = None

    def clone(self) -> "A2aAgent":
        """Clones the A2A agent."""
        import copy

        return A2aAgent(
            agent_card=copy.deepcopy(self.agent_card),
            task_store_builder=self._tmpl_attrs.get("task_store_builder"),
            task_store_kwargs=self._tmpl_attrs.get("task_store_kwargs"),
            agent_executor_kwargs=self._tmpl_attrs.get("agent_executor_kwargs"),
            agent_executor_builder=self._tmpl_attrs.get("agent_executor_builder"),
            request_handler_kwargs=self._tmpl_attrs.get("request_handler_kwargs"),
            request_handler_builder=self._tmpl_attrs.get("request_handler_builder"),
            extended_agent_card=self._tmpl_attrs.get("extended_agent_card"),
        )

    def set_up(self):
        """Sets up the A2A application."""
        # pylint: disable=g-import-not-at-top
        from a2a.server.apps.rest.rest_adapter import RESTAdapter
        from a2a.server.request_handlers.rest_handler import RESTHandler
        from a2a.server.request_handlers import DefaultRequestHandler
        from a2a.server.tasks import InMemoryTaskStore

        os.environ["GOOGLE_GENAI_USE_VERTEXAI"] = "1"
        project = self._tmpl_attrs.get("project")
        os.environ["GOOGLE_CLOUD_PROJECT"] = project
        location = self._tmpl_attrs.get("location")
        os.environ["GOOGLE_CLOUD_LOCATION"] = location
        agent_engine_id = os.getenv("GOOGLE_CLOUD_AGENT_ENGINE_ID", "test-agent-engine")
        version = "v1beta1"

        self.agent_card.url = f"https://{location}-aiplatform.googleapis.com/{version}/projects/{project}/locations/{location}/reasoningEngines/{agent_engine_id}/a2a"
        self._tmpl_attrs["agent_card"] = self.agent_card

        # Create the agent executor if a builder is provided.
        agent_executor_builder = self._tmpl_attrs.get("agent_executor_builder")
        if agent_executor_builder:
            self._tmpl_attrs["agent_executor"] = agent_executor_builder(
                **self._tmpl_attrs.get("agent_executor_kwargs")
            )
            self.agent_executor = self._tmpl_attrs.get("agent_executor")

        # Create the task store if a builder is provided.
        task_store_builder = self._tmpl_attrs.get("task_store_builder")
        if task_store_builder:
            self.task_store = task_store_builder(
                **self._tmpl_attrs.get("task_store_kwargs")
            )
        else:
            # Use the default task store if not provided. This could potentially
            # lead to unexpected behavior if the agent is running on
            # multiple instances.
            self.task_store = InMemoryTaskStore()

        self._tmpl_attrs["task_store"] = self.task_store

        # Create the request handler if a builder is provided.
        request_handler_builder = self._tmpl_attrs.get("request_handler_builder")
        if request_handler_builder:
            self.request_handler = request_handler_builder(
                **self._tmpl_attrs.get("request_handler_kwargs")
            )
        else:
            # Use the default request handler if not provided.
            self.request_handler = DefaultRequestHandler(
                agent_executor=self._tmpl_attrs.get("agent_executor"),
                task_store=self.task_store,
            )

        self._tmpl_attrs["request_handler"] = self.request_handler

        # a2a_rest_adapter is used to register the A2A API routes in the
        # Reasoning Engine API router.
        self.a2a_rest_adapter = RESTAdapter(
            agent_card=self.agent_card,
            http_handler=self._tmpl_attrs.get("request_handler"),
            extended_agent_card=self._tmpl_attrs.get("extended_agent_card"),
        )

        # rest_handler is used to handle the A2A API requests.
        self.rest_handler = RESTHandler(
            agent_card=self.agent_card,
            request_handler=self._tmpl_attrs.get("request_handler"),
        )

    async def on_message_send(
        self,
        request: "Request",
        context: "ServerCallContext",
    ) -> dict[str, Any]:
        return await self.rest_handler.on_message_send(request, context)

    async def on_cancel_task(
        self,
        request: "Request",
        context: "ServerCallContext",
    ) -> dict[str, Any]:
        return await self.rest_handler.on_cancel_task(request, context)

    async def on_get_task(
        self,
        request: "Request",
        context: "ServerCallContext",
    ) -> dict[str, Any]:
        return await self.rest_handler.on_get_task(request, context)

    async def handle_authenticated_agent_card(
        self,
        request: "Request",
        context: "ServerCallContext",
    ) -> dict[str, Any]:
        return await self.a2a_rest_adapter.handle_authenticated_agent_card(
            request, context
        )

    def register_operations(self) -> Dict[str, List[str]]:
        """Registers the operations of the A2A Agent."""
        routes = {
            "a2a_extension": [
                "on_message_send",
                "on_get_task",
                "on_cancel_task",
            ]
        }
        if self.agent_card.supports_authenticated_extended_card:
            routes["a2a_extension"].append("handle_authenticated_agent_card")
        return routes



"""
BMasterAI Enhanced Examples with Logging and Monitoring

This module contains comprehensive examples showing how to use BMasterAI
with the original stateful agent and multi-agent capabilities enhanced
with advanced logging and monitoring.
"""

import time
import threading
import json
from typing import Dict, List, Any, Optional
from dataclasses import dataclass
from datetime import datetime

# Import the new logging and monitoring systems
from bmasterai.logging import configure_logging, get_logger, LogLevel, EventType
from bmasterai.monitoring import get_monitor
from bmasterai.integrations import get_integration_manager, SlackConnector, EmailConnector


@dataclass
class AgentState:
    """Represents the state of an agent"""
    agent_id: str
    status: str
    current_task: Optional[str] = None
    task_count: int = 0
    last_activity: Optional[datetime] = None
    metadata: Dict[str, Any] = None

    def __post_init__(self):
        if self.metadata is None:
            self.metadata = {}
        if self.last_activity is None:
            self.last_activity = datetime.now()


class StatefulAgent:
    """
    Enhanced Stateful Agent with comprehensive logging and monitoring

    This agent maintains state across tasks and provides full observability
    into its operations through structured logging and metrics collection.
    """

    def __init__(self, agent_id: str, name: str, initial_state: Dict[str, Any] = None):
        self.agent_id = agent_id
        self.name = name
        self.state = AgentState(
            agent_id=agent_id,
            status="initialized",
            metadata=initial_state or {}
        )

        # Setup logging
        self.logger = get_logger()

        # Setup monitoring
        self.monitor = get_monitor()

        # Task history and state management
        self.task_history: List[Dict[str, Any]] = []
        self.memory: Dict[str, Any] = {}
        self.running = False
        self.thread: Optional[threading.Thread] = None

        # Log agent creation
        self.logger.log_event(
            agent_id,
            EventType.AGENT_START,
            f"Stateful agent {name} created",
            metadata={
                "agent_id": agent_id,
                "name": name,
                "initial_state": initial_state
            }
        )

    def start(self):
        """Start the agent with full monitoring"""
        if self.running:
            self.logger.log_event(self.agent_id, EventType.TASK_ERROR, "Agent already running")
            return

        self.running = True
        self.state.status = "running"
        self.state.last_activity = datetime.now()

        # Register with monitor
        self.monitor.track_agent_start(self.agent_id)

        # Start heartbeat thread
        self.thread = threading.Thread(target=self._heartbeat_loop, daemon=True)
        self.thread.start()

        self.logger.log_event(
            self.agent_id,
            EventType.AGENT_START,
            f"Stateful agent {self.name} started",
            metadata={"agent_id": self.agent_id}
        )

    def stop(self):
        """Stop the agent gracefully"""
        if not self.running:
            return

        self.running = False
        self.state.status = "stopped"

        # Track agent stop
        self.monitor.track_agent_stop(self.agent_id)

        if self.thread:
            self.thread.join(timeout=5)

        self.logger.log_event(
            self.agent_id,
            EventType.AGENT_STOP,
            f"Stateful agent {self.name} stopped",
            metadata={
                "agent_id": self.agent_id,
                "total_tasks": self.state.task_count,
                "final_state": self.get_state_summary()
            }
        )

    def _heartbeat_loop(self):
        """Send periodic heartbeats to monitoring system"""
        while self.running:
            try:
                # Record custom metric for agent status
                self.monitor.metrics_collector.record_custom_metric(
                    "agent_status",
                    1,
                    {
                        "agent_id": self.agent_id,
                        "status": self.state.status,
                        "task_count": str(self.state.task_count)
                    }
                )

                # Log heartbeat
                self.logger.log_event(
                    self.agent_id,
                    EventType.PERFORMANCE_METRIC,
                    f"Agent {self.name} heartbeat",
                    metadata={
                        "agent_id": self.agent_id,
                        "status": self.state.status,
                        "task_count": self.state.task_count
                    }
                )

                time.sleep(30)  # Heartbeat every 30 seconds

            except Exception as e:
                self.logger.log_event(
                    self.agent_id,
                    EventType.TASK_ERROR,
                    f"Heartbeat error for agent {self.name}",
                    metadata={"agent_id": self.agent_id, "error": str(e)}
                )
                time.sleep(60)  # Wait longer on error

    def execute_task(self, task_type: str, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Execute a task with full state management and logging"""
        if not self.running:
            error_msg = f"Agent {self.name} is not running"
            self.logger.log_event(EventType.TASK_ERROR, error_msg)
            return {"success": False, "error": error_msg}

        task_id = f"task_{int(time.time() * 1000)}"
        start_time = time.time()

        # Update state
        self.state.current_task = task_id
        self.state.last_activity = datetime.now()

        # Log task start
        self.logger.log_event(
            self.agent_id,
            EventType.TASK_START,
            f"Task {task_type} started",
            metadata={
                "agent_id": self.agent_id,
                "task_id": task_id,
                "task_type": task_type,
                "task_data": task_data
            }
        )

        try:
            # Execute the actual task based on type
            if task_type == "data_processing":
                result = self._process_data(task_data)
            elif task_type == "analysis":
                result = self._analyze_data(task_data)
            elif task_type == "memory_operation":
                result = self._memory_operation(task_data)
            elif task_type == "state_query":
                result = self._query_state(task_data)
            else:
                result = self._custom_task(task_type, task_data)

            # Update task count and history
            self.state.task_count += 1
            execution_time = time.time() - start_time

            task_record = {
                "task_id": task_id,
                "task_type": task_type,
                "start_time": start_time,
                "execution_time": execution_time,
                "success": result.get("success", True),
                "result_summary": str(result)[:200]  # Truncated for storage
            }

            self.task_history.append(task_record)

            # Keep only last 100 tasks in history
            if len(self.task_history) > 100:
                self.task_history = self.task_history[-100:]

            # Log successful completion
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_COMPLETE,
                f"Task {task_type} completed successfully",
                metadata={
                    "agent_id": self.agent_id,
                    "task_id": task_id,
                    "execution_time": execution_time,
                    "result": result
                }
            )

            # Record metrics
            self.monitor.track_task_duration(
                self.agent_id,
                task_type,
                execution_time * 1000  # Convert to milliseconds
            )

            return {
                "success": True,
                "task_id": task_id,
                "execution_time": execution_time,
                "result": result
            }

        except Exception as e:
            execution_time = time.time() - start_time
            error_msg = str(e)

            # Log error
            self.logger.log_event(
                self.agent_id,
                EventType.TASK_ERROR,
                f"Task {task_type} failed",
                metadata={
                    "agent_id": self.agent_id,
                    "task_id": task_id,
                    "error": error_msg,
                    "execution_time": execution_time
                }
            )

            # Record failed metrics
            self.monitor.track_error(self.agent_id, "task_failure")
            self.monitor.track_task_duration(
                self.agent_id,
                task_type,
                execution_time * 1000  # Convert to milliseconds
            )

            return {
                "success": False,
                "task_id": task_id,
                "error": error_msg,
                "execution_time": execution_time
            }

        finally:
            self.state.current_task = None

    def _process_data(self, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Process data and update agent memory"""
        data = task_data.get("data", [])
        operation = task_data.get("operation", "count")

        if operation == "count":
            result = len(data)
        elif operation == "sum" and all(isinstance(x, (int, float)) for x in data):
            result = sum(data)
        elif operation == "average" and all(isinstance(x, (int, float)) for x in data):
            result = sum(data) / len(data) if data else 0
        else:
            result = f"Processed {len(data)} items"

        # Store in memory
        memory_key = f"last_processing_{int(time.time())}"
        self.memory[memory_key] = {
            "operation": operation,
            "data_size": len(data),
            "result": result,
            "timestamp": datetime.now().isoformat()
        }

        return {
            "operation": operation,
            "result": result,
            "data_size": len(data),
            "memory_key": memory_key
        }

    def _analyze_data(self, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Analyze data using agent's memory and state"""
        analysis_type = task_data.get("type", "basic")
        data = task_data.get("data", {})

        # Use previous processing results from memory
        processing_history = [
            v for k, v in self.memory.items() 
            if k.startswith("last_processing_")
        ]

        analysis_result = {
            "analysis_type": analysis_type,
            "current_data": data,
            "historical_processing_count": len(processing_history),
            "agent_task_count": self.state.task_count,
            "analysis_timestamp": datetime.now().isoformat()
        }

        if analysis_type == "trend" and processing_history:
            # Simple trend analysis based on historical data
            recent_results = [h.get("result", 0) for h in processing_history[-5:]]
            if len(recent_results) > 1:
                trend = "increasing" if recent_results[-1] > recent_results[0] else "decreasing"
                analysis_result["trend"] = trend

        # Store analysis in memory
        memory_key = f"analysis_{int(time.time())}"
        self.memory[memory_key] = analysis_result

        return analysis_result

    def _memory_operation(self, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Perform operations on agent memory"""
        operation = task_data.get("operation", "list")

        if operation == "list":
            return {"memory_keys": list(self.memory.keys())}
        elif operation == "get":
            key = task_data.get("key")
            return {"key": key, "value": self.memory.get(key)}
        elif operation == "set":
            key = task_data.get("key")
            value = task_data.get("value")
            self.memory[key] = value
            return {"key": key, "set": True}
        elif operation == "clear":
            cleared_count = len(self.memory)
            self.memory.clear()
            return {"cleared_items": cleared_count}
        elif operation == "size":
            return {"memory_size": len(self.memory)}
        else:
            return {"error": f"Unknown memory operation: {operation}"}

    def _query_state(self, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Query agent state information"""
        return self.get_state_summary()

    def _custom_task(self, task_type: str, task_data: Dict[str, Any]) -> Dict[str, Any]:
        """Handle custom task types - override in subclasses"""
        return {
            "message": f"Custom task {task_type} executed",
            "task_data": task_data,
            "agent_state": self.get_state_summary()
        }

    def get_state_summary(self) -> Dict[str, Any]:
        """Get a summary of the agent's current state"""
        return {
            "agent_id": self.agent_id,
            "name": self.name,
            "status": self.state.status,
            "task_count": self.state.task_count,
            "current_task": self.state.current_task,
            "memory_size": len(self.memory),
            "last_activity": self.state.last_activity.isoformat() if self.state.last_activity else None,
            "recent_tasks": self.task_history[-5:] if self.task_history else []
        }


class MultiAgentOrchestrator:
    """
    Enhanced Multi-Agent Orchestrator with comprehensive logging and monitoring

    Coordinates multiple stateful agents and provides centralized management
    with full observability into multi-agent operations.
    """

    def __init__(self, orchestrator_id: str = "orchestrator"):
        self.orchestrator_id = orchestrator_id
        self.agents: Dict[str, StatefulAgent] = {}
        self.task_queue: List[Dict[str, Any]] = []
        self.coordination_history: List[Dict[str, Any]] = []

        # Setup logging and monitoring
        self.logger = get_logger()
        self.monitor = get_monitor()

        # Integration manager for notifications
        self.integration_manager = get_integration_manager()

        self.logger.log_event(
            orchestrator_id,
            EventType.AGENT_START,
            f"Multi-agent orchestrator {orchestrator_id} initialized",
            metadata={"orchestrator_id": orchestrator_id}
        )

    def add_agent(self, agent: StatefulAgent):
        """Add an agent to the orchestrator"""
        self.agents[agent.agent_id] = agent

        self.logger.log_event(
            self.orchestrator_id,
            EventType.AGENT_COMMUNICATION,
            f"Agent {agent.name} added to orchestrator",
            metadata={
                "orchestrator_id": self.orchestrator_id,
                "agent_id": agent.agent_id,
                "agent_name": agent.name
            }
        )

    def remove_agent(self, agent_id: str):
        """Remove an agent from the orchestrator"""
        if agent_id in self.agents:
            agent = self.agents[agent_id]
            agent.stop()
            del self.agents[agent_id]

            self.logger.log_event(
                self.orchestrator_id,
                EventType.AGENT_COMMUNICATION,
                f"Agent {agent_id} removed from orchestrator",
                metadata={
                    "orchestrator_id": self.orchestrator_id,
                    "agent_id": agent_id
                }
            )

    def start_all_agents(self):
        """Start all registered agents"""
        started_count = 0
        for agent in self.agents.values():
            try:
                agent.start()
                started_count += 1
            except Exception as e:
                self.logger.log_event(
                    EventType.AGENT_ERROR,
                    f"Failed to start agent {agent.agent_id}",
                    {"error": str(e)}
                )

        self.logger.log_event(
            self.orchestrator_id,
            EventType.AGENT_START,
            f"Started {started_count}/{len(self.agents)} agents",
            metadata={
                "orchestrator_id": self.orchestrator_id,
                "started_count": started_count,
                "total_agents": len(self.agents)
            }
        )

        return started_count

    def stop_all_agents(self):
        """Stop all agents gracefully"""
        stopped_count = 0
        for agent in self.agents.values():
            try:
                agent.stop()
                stopped_count += 1
            except Exception as e:
                self.logger.log_event(
                    EventType.AGENT_ERROR,
                    f"Failed to stop agent {agent.agent_id}",
                    {"error": str(e)}
                )

        self.logger.log_event(
            self.orchestrator_id,
            EventType.AGENT_STOP,
            f"Stopped {stopped_count}/{len(self.agents)} agents",
            metadata={
                "orchestrator_id": self.orchestrator_id,
                "stopped_count": stopped_count
            }
        )

        return stopped_count

    def coordinate_task(self, coordination_id: str, task_assignments: Dict[str, tuple]) -> Dict[str, Any]:
        """
        Coordinate a complex task across multiple agents

        Args:
            coordination_id: Unique identifier for this coordination
            task_assignments: Dict mapping agent_id to (task_type, task_data) tuples
        """
        start_time = time.time()

        self.logger.log_event(
            self.orchestrator_id,
            EventType.TASK_START,
            f"Multi-agent coordination {coordination_id} started",
            metadata={
                "orchestrator_id": self.orchestrator_id,
                "coordination_id": coordination_id,
                "agent_assignments": list(task_assignments.keys()),
                "total_tasks": len(task_assignments)
            }
        )

        results = {}
        errors = {}

        # Execute tasks on assigned agents
        for agent_id, (task_type, task_data) in task_assignments.items():
            if agent_id not in self.agents:
                error_msg = f"Agent {agent_id} not found in orchestrator"
                errors[agent_id] = error_msg
                self.logger.log_event(
                    EventType.TASK_ERROR,
                    error_msg,
                    {"coordination_id": coordination_id}
                )
                continue

            try:
                agent = self.agents[agent_id]
                result = agent.execute_task(task_type, task_data)
                results[agent_id] = result

                if not result.get("success", False):
                    errors[agent_id] = result.get("error", "Task failed")

            except Exception as e:
                error_msg = str(e)
                errors[agent_id] = error_msg
                self.logger.log_event(
                    EventType.TASK_ERROR,
                    f"Error executing task on agent {agent_id}",
                    {
                        "coordination_id": coordination_id,
                        "agent_id": agent_id,
                        "error": error_msg
                    }
                )

        execution_time = time.time() - start_time
        success = len(errors) == 0

        # Record coordination history
        coordination_record = {
            "coordination_id": coordination_id,
            "start_time": start_time,
            "execution_time": execution_time,
            "success": success,
            "agents_involved": list(task_assignments.keys()),
            "successful_agents": [aid for aid in results.keys() if aid not in errors],
            "failed_agents": list(errors.keys()),
            "timestamp": datetime.now().isoformat()
        }

        self.coordination_history.append(coordination_record)

        # Keep only last 50 coordination records
        if len(self.coordination_history) > 50:
            self.coordination_history = self.coordination_history[-50:]

        # Log completion
        if success:
            self.logger.log_event(
                self.orchestrator_id,
                EventType.TASK_COMPLETE,
                f"Multi-agent coordination {coordination_id} completed successfully",
                metadata={
                    "coordination_id": coordination_id,
                    "execution_time": execution_time,
                    "agents_count": len(results)
                }
            )

            # Send success notification
            self._send_coordination_notification(coordination_id, True, execution_time, len(results))

        else:
            self.logger.log_event(
                EventType.TASK_ERROR,
                f"Multi-agent coordination {coordination_id} completed with errors",
                {
                    "coordination_id": coordination_id,
                    "execution_time": execution_time,
                    "errors": errors
                }
            )

            # Send error notification
            self._send_coordination_notification(coordination_id, False, execution_time, len(results), errors)

        return {
            "coordination_id": coordination_id,
            "success": success,
            "execution_time": execution_time,
            "results": results,
            "errors": errors,
            "summary": {
                "total_agents": len(task_assignments),
                "successful_agents": len(results) - len(errors),
                "failed_agents": len(errors)
            }
        }

    def _send_coordination_notification(self, coordination_id: str, success: bool, 
                                      execution_time: float, agent_count: int, 
                                      errors: Dict[str, str] = None):
        """Send notification about coordination completion"""
        try:
            status = "✅ SUCCESS" if success else "❌ FAILED"
            message = f"{status} Multi-Agent Coordination: {coordination_id}\n"
            message += f"⏱️ Execution Time: {execution_time:.2f}s\n"
            message += f"🤖 Agents Involved: {agent_count}\n"

            if errors:
                message += f"❌ Errors: {len(errors)}\n"
                for agent_id, error in errors.items():
                    message += f"  • {agent_id}: {error}\n"

            # Send to all configured notification channels
            self.integration_manager.send_notification(message)

        except Exception as e:
            self.logger.log_event(
                self.orchestrator_id,
                EventType.TASK_ERROR,
                f"Failed to send coordination notification",
                metadata={"error": str(e)}
            )

    def get_orchestrator_status(self) -> Dict[str, Any]:
        """Get comprehensive orchestrator status"""
        agent_statuses = {}
        for agent_id, agent in self.agents.items():
            agent_statuses[agent_id] = agent.get_state_summary()

        return {
            "orchestrator_id": self.orchestrator_id,
            "total_agents": len(self.agents),
            "running_agents": sum(1 for a in self.agents.values() if a.running),
            "total_coordinations": len(self.coordination_history),
            "recent_coordinations": self.coordination_history[-5:],
            "agent_statuses": agent_statuses,
            "system_health": self.monitor.get_system_health()
        }

    def get_agent_performance_report(self) -> Dict[str, Any]:
        """Generate performance report for all agents"""
        report = {
            "orchestrator_id": self.orchestrator_id,
            "report_timestamp": datetime.now().isoformat(),
            "agents": {}
        }

        for agent_id, agent in self.agents.items():
            agent_report = {
                "agent_summary": agent.get_state_summary(),
                "performance_metrics": self.monitor.get_agent_dashboard(agent_id)
            }
            report["agents"][agent_id] = agent_report

        return report


# Example usage functions
def example_stateful_agent_with_logging():
    """Example: Stateful agent with comprehensive logging"""
    print("🚀 Starting Stateful Agent Example with Logging")

    # Configure logging
    logger = configure_logging(log_level=LogLevel.INFO)

    # Start monitoring
    monitor = get_monitor()
    monitor.start_monitoring()

    # Setup integrations (optional)
    integration_manager = get_integration_manager()
    # Uncomment to add Slack notifications:
    # slack = SlackConnector(webhook_url="YOUR_WEBHOOK_URL")
    # integration_manager.add_connector("slack", slack)

    # Create and start agent
    agent = StatefulAgent("data-processor-001", "DataProcessor", {"version": "1.0"})
    agent.start()

    try:
        # Execute various tasks to demonstrate state management
        print("\n📊 Executing data processing tasks...")

        # Task 1: Process some data
        result1 = agent.execute_task("data_processing", {
            "data": [1, 2, 3, 4, 5],
            "operation": "sum"
        })
        print(f"Task 1 Result: {result1}")

        # Task 2: Analyze trends
        result2 = agent.execute_task("analysis", {
            "type": "trend",
            "data": {"current_value": 15}
        })
        print(f"Task 2 Result: {result2}")

        # Task 3: Memory operations
        result3 = agent.execute_task("memory_operation", {
            "operation": "set",
            "key": "user_preference",
            "value": {"theme": "dark", "notifications": True}
        })
        print(f"Task 3 Result: {result3}")

        # Task 4: Query state
        result4 = agent.execute_task("state_query", {})
        print(f"Task 4 Result: {result4}")

        # Show agent state
        print(f"\n🔍 Agent State Summary:")
        print(json.dumps(agent.get_state_summary(), indent=2))

        # Show performance metrics
        print(f"\n📈 Performance Dashboard:")
        dashboard = monitor.get_agent_dashboard(agent.agent_id)
        print(json.dumps(dashboard, indent=2))

    finally:
        agent.stop()
        monitor.stop_monitoring()
        print("\n✅ Stateful agent example completed")


def example_multi_agent_coordination_with_logging():
    """Example: Multi-agent coordination with comprehensive logging"""
    print("🚀 Starting Multi-Agent Coordination Example with Logging")

    # Configure logging
    logger = configure_logging(log_level=LogLevel.INFO)

    # Start monitoring
    monitor = get_monitor()
    monitor.start_monitoring()

    # Create orchestrator
    orchestrator = MultiAgentOrchestrator("main-orchestrator")

    # Create specialized agents
    data_agent = StatefulAgent("data-agent", "DataProcessor")
    analysis_agent = StatefulAgent("analysis-agent", "DataAnalyzer") 
    report_agent = StatefulAgent("report-agent", "ReportGenerator")

    # Add agents to orchestrator
    orchestrator.add_agent(data_agent)
    orchestrator.add_agent(analysis_agent)
    orchestrator.add_agent(report_agent)

    try:
        # Start all agents
        started_count = orchestrator.start_all_agents()
        print(f"✅ Started {started_count} agents")

        # Wait a moment for agents to initialize
        time.sleep(2)

        # Coordinate a complex multi-agent task
        print("\n🤝 Coordinating multi-agent task...")

        task_assignments = {
            "data-agent": ("data_processing", {
                "data": [10, 20, 30, 40, 50],
                "operation": "average"
            }),
            "analysis-agent": ("analysis", {
                "type": "trend",
                "data": {"period": "daily", "metric": "sales"}
            }),
            "report-agent": ("memory_operation", {
                "operation": "set",
                "key": "report_config",
                "value": {"format": "pdf", "include_charts": True}
            })
        }

        coordination_result = orchestrator.coordinate_task(
            "daily_analysis_workflow",
            task_assignments
        )

        print(f"\n📋 Coordination Result:")
        print(json.dumps(coordination_result, indent=2))

        # Show orchestrator status
        print(f"\n🎛️ Orchestrator Status:")
        status = orchestrator.get_orchestrator_status()
        print(json.dumps(status, indent=2))

        # Generate performance report
        print(f"\n📊 Performance Report:")
        performance_report = orchestrator.get_agent_performance_report()
        print(json.dumps(performance_report, indent=2))

    finally:
        orchestrator.stop_all_agents()
        monitor.stop_monitoring()
        print("\n✅ Multi-agent coordination example completed")


def example_advanced_monitoring_and_alerts():
    """Example: Advanced monitoring with custom alerts"""
    print("🚀 Starting Advanced Monitoring Example")

    # Configure logging with JSON output
    logger = configure_logging(
        log_level=LogLevel.INFO,
        enable_json=True,
        enable_file=True
    )

    # Start monitoring with custom configuration
    monitor = get_monitor()
    monitor.start_monitoring()

    # Setup custom alerts
    monitor.add_alert_rule({
        "metric_name": "agent_task_failure_rate",
        "threshold": 0.5,  # 50% failure rate
        "condition": "greater_than",
        "duration_minutes": 2,
        "callback": lambda alert: print(f"🚨 ALERT: {alert}")
    })

    # Create agents with different behaviors
    reliable_agent = StatefulAgent("reliable-agent", "ReliableAgent")
    unreliable_agent = StatefulAgent("unreliable-agent", "UnreliableAgent")

    # Custom unreliable agent that fails sometimes
    class UnreliableAgent(StatefulAgent):
        def _custom_task(self, task_type: str, task_data: Dict[str, Any]) -> Dict[str, Any]:
            import random
            if random.random() < 0.6:  # 60% chance of failure
                raise Exception("Simulated task failure")
            return {"message": "Task succeeded"}

    unreliable_agent.__class__ = UnreliableAgent

    try:
        reliable_agent.start()
        unreliable_agent.start()

        # Execute tasks to trigger monitoring
        print("\n📊 Executing tasks to demonstrate monitoring...")

        for i in range(10):
            # Reliable agent tasks
            reliable_agent.execute_task("data_processing", {
                "data": list(range(i, i+5)),
                "operation": "sum"
            })

            # Unreliable agent tasks (will trigger alerts)
            unreliable_agent.execute_task("custom_task", {"iteration": i})

            time.sleep(1)

        # Wait for monitoring to collect data
        time.sleep(5)

        # Show monitoring results
        print(f"\n📈 System Health:")
        health = monitor.get_system_health()
        print(json.dumps(health, indent=2))

        print(f"\n📊 Agent Dashboards:")
        for agent_id in [reliable_agent.agent_id, unreliable_agent.agent_id]:
            dashboard = monitor.get_agent_dashboard(agent_id)
            print(f"\n{agent_id}:")
            print(json.dumps(dashboard, indent=2))

    finally:
        reliable_agent.stop()
        unreliable_agent.stop()
        monitor.stop_monitoring()
        print("\n✅ Advanced monitoring example completed")


if __name__ == "__main__":
    print("BMasterAI Enhanced Examples")
    print("=" * 50)

    # Run examples
    example_stateful_agent_with_logging()
    print("\n" + "="*50 + "\n")

    example_multi_agent_coordination_with_logging()
    print("\n" + "="*50 + "\n")

    example_advanced_monitoring_and_alerts()

    print("\n🎉 All examples completed successfully!")



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Classes for supervised tuning."""

# We just want to re-export certain classes
# pylint: disable=g-multiple-import,g-importing-member
from vertexai.preview.tuning._supervised_tuning import (
    train as preview_train,
    SupervisedTuningJob as PreviewSupervisedTuningJob,
)
from vertexai.tuning._supervised_tuning import (
    train,
    SupervisedTuningJob,
)
from vertexai.tuning._tuning import (
    rebase_tuned_model,
)

__all__ = [
    "rebase_tuned_model",
    "train",
    "SupervisedTuningJob",
    "PreviewSupervisedTuningJob",
    "preview_train",
]



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore
from google.type import interval_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "FeatureMonitor",
        "ScheduleConfig",
        "FeatureSelectionConfig",
        "FeatureStatsAndAnomaly",
        "FeatureStatsAndAnomalySpec",
    },
)


class FeatureMonitor(proto.Message):
    r"""Vertex AI Feature Monitor.

    Attributes:
        name (str):
            Identifier. Name of the FeatureMonitor. Format:
            ``projects/{project}/locations/{location}/featureGroups/{featureGroup}/featureMonitors/{featureMonitor}``
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            FeatureMonitor was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Timestamp when this
            FeatureMonitor was last updated.
        etag (str):
            Optional. Used to perform consistent
            read-modify-write updates. If not set, a blind
            "overwrite" update happens.
        labels (MutableMapping[str, str]):
            Optional. The labels with user-defined
            metadata to organize your FeatureMonitor.

            Label keys and values can be no longer than 64
            characters (Unicode codepoints), can only
            contain lowercase letters, numeric characters,
            underscores and dashes. International characters
            are allowed.

            See https://goo.gl/xmQnxf for more information
            on and examples of labels. No more than 64 user
            labels can be associated with one
            FeatureMonitor(System labels are excluded)."
            System reserved label keys are prefixed with
            "aiplatform.googleapis.com/" and are immutable.
        description (str):
            Optional. Description of the FeatureMonitor.
        schedule_config (google.cloud.aiplatform_v1beta1.types.ScheduleConfig):
            Required. Schedule config for the
            FeatureMonitor.
        feature_selection_config (google.cloud.aiplatform_v1beta1.types.FeatureSelectionConfig):
            Required. Feature selection config for the
            FeatureMonitor.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    etag: str = proto.Field(
        proto.STRING,
        number=4,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=5,
    )
    description: str = proto.Field(
        proto.STRING,
        number=6,
    )
    schedule_config: "ScheduleConfig" = proto.Field(
        proto.MESSAGE,
        number=7,
        message="ScheduleConfig",
    )
    feature_selection_config: "FeatureSelectionConfig" = proto.Field(
        proto.MESSAGE,
        number=8,
        message="FeatureSelectionConfig",
    )


class ScheduleConfig(proto.Message):
    r"""Schedule configuration for the FeatureMonitor.

    Attributes:
        cron (str):
            Cron schedule (https://en.wikipedia.org/wiki/Cron) to launch
            scheduled runs. To explicitly set a timezone to the cron
            tab, apply a prefix in the cron tab:
            "CRON_TZ=${IANA_TIME_ZONE}" or "TZ=${IANA_TIME_ZONE}". The
            ${IANA_TIME_ZONE} may only be a valid string from IANA time
            zone database. For example, "CRON_TZ=America/New_York 1 \*
            \* \* \*", or "TZ=America/New_York 1 \* \* \* \*".
    """

    cron: str = proto.Field(
        proto.STRING,
        number=1,
    )


class FeatureSelectionConfig(proto.Message):
    r"""Feature selection configuration for the FeatureMonitor.

    Attributes:
        feature_configs (MutableSequence[google.cloud.aiplatform_v1beta1.types.FeatureSelectionConfig.FeatureConfig]):
            Optional. A list of features to be monitored
            and each feature's drift threshold.
    """

    class FeatureConfig(proto.Message):
        r"""Feature configuration.

        Attributes:
            feature_id (str):
                Required. The ID of the feature resource.
                Final component of the Feature's resource name.
            drift_threshold (float):
                Optional. Drift threshold. If calculated
                difference with baseline data larger than
                threshold, it will be considered as the feature
                has drift. If not present, the threshold will be
                default to 0.3.
        """

        feature_id: str = proto.Field(
            proto.STRING,
            number=1,
        )
        drift_threshold: float = proto.Field(
            proto.DOUBLE,
            number=2,
        )

    feature_configs: MutableSequence[FeatureConfig] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message=FeatureConfig,
    )


class FeatureStatsAndAnomaly(proto.Message):
    r"""Stats and Anomaly generated by FeatureMonitorJobs. Anomaly
    only includes Drift.

    Attributes:
        feature_id (str):
            Feature Id.
        feature_stats (google.protobuf.struct_pb2.Value):
            Feature stats. e.g. histogram buckets.
            In the format of
            tensorflow.metadata.v0.DatasetFeatureStatistics.
        distribution_deviation (float):
            Deviation from the current stats to baseline
            stats.
              1. For categorical feature, the distribution
                distance is calculated by      L-inifinity
                norm.
              2. For numerical feature, the distribution
                distance is calculated by
                Jensen–Shannon divergence.
        drift_detection_threshold (float):
            This is the threshold used when detecting drifts, which is
            set in
            FeatureMonitor.FeatureSelectionConfig.FeatureConfig.drift_threshold
        drift_detected (bool):
            If set to true, indicates current stats is
            detected as and comparing with baseline stats.
        stats_time (google.protobuf.timestamp_pb2.Timestamp):
            The timestamp we take snapshot for feature
            values to generate stats.
        feature_monitor_job_id (int):
            The ID of the FeatureMonitorJob that
            generated this FeatureStatsAndAnomaly.
        feature_monitor_id (str):
            The ID of the FeatureMonitor that this
            FeatureStatsAndAnomaly generated according to.
    """

    feature_id: str = proto.Field(
        proto.STRING,
        number=1,
    )
    feature_stats: struct_pb2.Value = proto.Field(
        proto.MESSAGE,
        number=2,
        message=struct_pb2.Value,
    )
    distribution_deviation: float = proto.Field(
        proto.DOUBLE,
        number=3,
    )
    drift_detection_threshold: float = proto.Field(
        proto.DOUBLE,
        number=4,
    )
    drift_detected: bool = proto.Field(
        proto.BOOL,
        number=5,
    )
    stats_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=6,
        message=timestamp_pb2.Timestamp,
    )
    feature_monitor_job_id: int = proto.Field(
        proto.INT64,
        number=7,
    )
    feature_monitor_id: str = proto.Field(
        proto.STRING,
        number=8,
    )


class FeatureStatsAndAnomalySpec(proto.Message):
    r"""Defines how to select FeatureStatsAndAnomaly to be populated
    in response. If set, retrieves FeatureStatsAndAnomaly generated
    by FeatureMonitors based on this spec.


    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        latest_stats_count (int):
            Optional. If set, returns the most recent count of stats.
            Valid value is [0, 100]. If stats_time_range is set, return
            most recent count of stats within the stats_time_range.

            This field is a member of `oneof`_ ``_latest_stats_count``.
        stats_time_range (google.type.interval_pb2.Interval):
            Optional. If set, return all stats generated between
            [start_time, end_time). If latest_stats_count is set, return
            the most recent count of stats within the stats_time_range.
    """

    latest_stats_count: int = proto.Field(
        proto.INT32,
        number=1,
        optional=True,
    )
    stats_time_range: interval_pb2.Interval = proto.Field(
        proto.MESSAGE,
        number=2,
        message=interval_pb2.Interval,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from google.cloud.bigtable.data.execute_query._checksum import _CRC32C
from google.cloud.bigtable_v2 import ExecuteQueryResponse


class _ByteCursor:
    """
    Buffers bytes from `ExecuteQuery` responses until resume_token is received or end-of-stream
    is reached. :class:`google.cloud.bigtable_v2.types.bigtable.ExecuteQueryResponse` obtained from
    the server should be passed to the ``consume`` method and its non-None results should be passed
    to appropriate :class:`google.cloud.bigtable.execute_query_reader._Reader` for parsing gathered
    bytes.

    This class consumes data obtained externally to be usable in both sync and async clients.

    See :class:`google.cloud.bigtable.execute_query_reader._Reader` for more context.
    """

    def __init__(self):
        self._batch_buffer = bytearray()
        self._batches: List[bytes] = []
        self._resume_token = None

    def reset(self):
        self._batch_buffer = bytearray()
        self._batches = []

    def prepare_for_new_request(self):
        """
        Prepares this ``_ByteCursor`` for retrying an ``ExecuteQuery`` request.

        Clears internal buffers of this ``_ByteCursor`` and returns last received
        ``resume_token`` to be used in retried request.

        This is the only method that returns ``resume_token`` to the user.
        Returning the token to the user is tightly coupled with clearing internal
        buffers to prevent accidental retry without clearing the state, what would
        cause invalid results. ``resume_token`` are not needed in other cases,
        thus they is no separate getter for it.

        Returns:
            bytes: Last received resume_token.
        """
        # The first response of any retried stream will always contain reset, so
        # this isn't actually necessary, but we do it for safety
        self.reset()
        return self._resume_token

    def empty(self) -> bool:
        return not self._batch_buffer and not self._batches

    def consume(self, response: ExecuteQueryResponse) -> Optional[List[bytes]]:
        """
        Reads results bytes from an ``ExecuteQuery`` response and adds them to a buffer.

        If the response contains a ``resume_token``:
        - the ``resume_token`` is saved in this ``_ByteCursor``, and
        - internal buffers are flushed and returned to the caller.

        ``resume_token`` is not available directly, but can be retrieved by calling
        :meth:`._ByteCursor.prepare_for_new_request` when preparing to retry a request.

        Args:
            response (google.cloud.bigtable_v2.types.bigtable.ExecuteQueryResponse):
                Response obtained from the stream.

        Returns:
            bytes or None: List of bytes if buffers were flushed or None otherwise.
            Each element in the list represents the bytes of a `ProtoRows` message.

        Raises:
            ValueError: If provided ``ExecuteQueryResponse`` is not valid
                or contains bytes representing response of a different kind than previously
                processed responses.
        """
        response_pb = response._pb  # proto-plus attribute retrieval is slow.

        if response_pb.HasField("results"):
            results = response_pb.results
            if results.reset:
                self.reset()
            if results.HasField("proto_rows_batch"):
                self._batch_buffer.extend(results.proto_rows_batch.batch_data)
                # Note that 0 is a valid checksum so we must check for field presence
                if results.HasField("batch_checksum"):
                    expected_checksum = results.batch_checksum
                    checksum = _CRC32C.checksum(self._batch_buffer)
                    if expected_checksum != checksum:
                        raise ValueError(
                            f"Unexpected checksum mismatch. Expected: {expected_checksum}, got: {checksum}"
                        )
                    # We have a complete batch so we move it to batches and reset the
                    # batch_buffer
                    self._batches.append(memoryview(self._batch_buffer))
                    self._batch_buffer = bytearray()

            if results.resume_token:
                self._resume_token = results.resume_token

                if self._batches:
                    if self._batch_buffer:
                        raise ValueError("Unexpected resume_token without checksum")
                    return_value = self._batches
                    self._batches = []
                    return return_value
        else:
            raise ValueError(f"Unexpected ExecuteQueryResponse: {response}")
        return None



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import openapi
from google.protobuf import struct_pb2  # type: ignore
from google.type import latlng_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "Tool",
        "UrlContext",
        "FunctionDeclaration",
        "FunctionCall",
        "PartialArg",
        "FunctionResponsePart",
        "FunctionResponseBlob",
        "FunctionResponseFileData",
        "FunctionResponse",
        "ExecutableCode",
        "CodeExecutionResult",
        "Retrieval",
        "VertexRagStore",
        "VertexAISearch",
        "GoogleSearchRetrieval",
        "GoogleMaps",
        "EnterpriseWebSearch",
        "DynamicRetrievalConfig",
        "ToolConfig",
        "FunctionCallingConfig",
        "RetrievalConfig",
        "RagRetrievalConfig",
    },
)


class Tool(proto.Message):
    r"""Tool details that the model may use to generate response.

    A ``Tool`` is a piece of code that enables the system to interact
    with external systems to perform an action, or set of actions,
    outside of knowledge and scope of the model. A Tool object should
    contain exactly one type of Tool (e.g FunctionDeclaration, Retrieval
    or GoogleSearchRetrieval).

    Attributes:
        function_declarations (MutableSequence[google.cloud.aiplatform_v1.types.FunctionDeclaration]):
            Optional. Function tool type. One or more function
            declarations to be passed to the model along with the
            current user query. Model may decide to call a subset of
            these functions by populating
            [FunctionCall][google.cloud.aiplatform.v1.Part.function_call]
            in the response. User should provide a
            [FunctionResponse][google.cloud.aiplatform.v1.Part.function_response]
            for each function call in the next turn. Based on the
            function responses, Model will generate the final response
            back to the user. Maximum 128 function declarations can be
            provided.
        retrieval (google.cloud.aiplatform_v1.types.Retrieval):
            Optional. Retrieval tool type.
            System will always execute the provided
            retrieval tool(s) to get external knowledge to
            answer the prompt. Retrieval results are
            presented to the model for generation.
        google_search (google.cloud.aiplatform_v1.types.Tool.GoogleSearch):
            Optional. GoogleSearch tool type.
            Tool to support Google Search in Model. Powered
            by Google.
        google_search_retrieval (google.cloud.aiplatform_v1.types.GoogleSearchRetrieval):
            Optional. GoogleSearchRetrieval tool type.
            Specialized retrieval tool that is powered by
            Google search.
        google_maps (google.cloud.aiplatform_v1.types.GoogleMaps):
            Optional. GoogleMaps tool type.
            Tool to support Google Maps in Model.
        enterprise_web_search (google.cloud.aiplatform_v1.types.EnterpriseWebSearch):
            Optional. Tool to support searching public
            web data, powered by Vertex AI Search and Sec4
            compliance.
        code_execution (google.cloud.aiplatform_v1.types.Tool.CodeExecution):
            Optional. CodeExecution tool type.
            Enables the model to execute code as part of
            generation.
        url_context (google.cloud.aiplatform_v1.types.UrlContext):
            Optional. Tool to support URL context
            retrieval.
        computer_use (google.cloud.aiplatform_v1.types.Tool.ComputerUse):
            Optional. Tool to support the model
            interacting directly with the computer. If
            enabled, it automatically populates computer-use
            specific Function Declarations.
    """

    class PhishBlockThreshold(proto.Enum):
        r"""These are available confidence level user can set to block
        malicious urls with chosen confidence and above. For
        understanding different confidence of webrisk, please refer to
        https://cloud.google.com/web-risk/docs/reference/rpc/google.cloud.webrisk.v1eap1#confidencelevel

        Values:
            PHISH_BLOCK_THRESHOLD_UNSPECIFIED (0):
                Defaults to unspecified.
            BLOCK_LOW_AND_ABOVE (30):
                Blocks Low and above confidence URL that is
                risky.
            BLOCK_MEDIUM_AND_ABOVE (40):
                Blocks Medium and above confidence URL that
                is risky.
            BLOCK_HIGH_AND_ABOVE (50):
                Blocks High and above confidence URL that is
                risky.
            BLOCK_HIGHER_AND_ABOVE (55):
                Blocks Higher and above confidence URL that
                is risky.
            BLOCK_VERY_HIGH_AND_ABOVE (60):
                Blocks Very high and above confidence URL
                that is risky.
            BLOCK_ONLY_EXTREMELY_HIGH (100):
                Blocks Extremely high confidence URL that is
                risky.
        """

        PHISH_BLOCK_THRESHOLD_UNSPECIFIED = 0
        BLOCK_LOW_AND_ABOVE = 30
        BLOCK_MEDIUM_AND_ABOVE = 40
        BLOCK_HIGH_AND_ABOVE = 50
        BLOCK_HIGHER_AND_ABOVE = 55
        BLOCK_VERY_HIGH_AND_ABOVE = 60
        BLOCK_ONLY_EXTREMELY_HIGH = 100

    class GoogleSearch(proto.Message):
        r"""GoogleSearch tool type.
        Tool to support Google Search in Model. Powered by Google.


        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            exclude_domains (MutableSequence[str]):
                Optional. List of domains to be excluded from the search
                results. The default limit is 2000 domains. Example:
                ["amazon.com", "facebook.com"].
            blocking_confidence (google.cloud.aiplatform_v1.types.Tool.PhishBlockThreshold):
                Optional. Sites with confidence level chosen
                & above this value will be blocked from the
                search results.

                This field is a member of `oneof`_ ``_blocking_confidence``.
        """

        exclude_domains: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=3,
        )
        blocking_confidence: "Tool.PhishBlockThreshold" = proto.Field(
            proto.ENUM,
            number=4,
            optional=True,
            enum="Tool.PhishBlockThreshold",
        )

    class CodeExecution(proto.Message):
        r"""Tool that executes code generated by the model, and automatically
        returns the result to the model.

        See also [ExecutableCode]and [CodeExecutionResult] which are input
        and output to this tool.

        """

    class ComputerUse(proto.Message):
        r"""Tool to support computer use.

        Attributes:
            environment (google.cloud.aiplatform_v1.types.Tool.ComputerUse.Environment):
                Required. The environment being operated.
            excluded_predefined_functions (MutableSequence[str]):
                Optional. By default, `predefined
                functions <https://cloud.google.com/vertex-ai/generative-ai/docs/computer-use#supported-actions>`__
                are included in the final model call. Some of them can be
                explicitly excluded from being automatically included. This
                can serve two purposes:

                1. Using a more restricted / different action space.
                2. Improving the definitions / instructions of predefined
                   functions.
        """

        class Environment(proto.Enum):
            r"""Represents the environment being operated, such as a web
            browser.

            Values:
                ENVIRONMENT_UNSPECIFIED (0):
                    Defaults to browser.
                ENVIRONMENT_BROWSER (1):
                    Operates in a web browser.
            """

            ENVIRONMENT_UNSPECIFIED = 0
            ENVIRONMENT_BROWSER = 1

        environment: "Tool.ComputerUse.Environment" = proto.Field(
            proto.ENUM,
            number=1,
            enum="Tool.ComputerUse.Environment",
        )
        excluded_predefined_functions: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=2,
        )

    function_declarations: MutableSequence["FunctionDeclaration"] = proto.RepeatedField(
        proto.MESSAGE,
        number=1,
        message="FunctionDeclaration",
    )
    retrieval: "Retrieval" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="Retrieval",
    )
    google_search: GoogleSearch = proto.Field(
        proto.MESSAGE,
        number=7,
        message=GoogleSearch,
    )
    google_search_retrieval: "GoogleSearchRetrieval" = proto.Field(
        proto.MESSAGE,
        number=3,
        message="GoogleSearchRetrieval",
    )
    google_maps: "GoogleMaps" = proto.Field(
        proto.MESSAGE,
        number=5,
        message="GoogleMaps",
    )
    enterprise_web_search: "EnterpriseWebSearch" = proto.Field(
        proto.MESSAGE,
        number=6,
        message="EnterpriseWebSearch",
    )
    code_execution: CodeExecution = proto.Field(
        proto.MESSAGE,
        number=4,
        message=CodeExecution,
    )
    url_context: "UrlContext" = proto.Field(
        proto.MESSAGE,
        number=10,
        message="UrlContext",
    )
    computer_use: ComputerUse = proto.Field(
        proto.MESSAGE,
        number=11,
        message=ComputerUse,
    )


class UrlContext(proto.Message):
    r"""Tool to support URL context."""


class FunctionDeclaration(proto.Message):
    r"""Structured representation of a function declaration as defined by
    the `OpenAPI 3.0
    specification <https://spec.openapis.org/oas/v3.0.3>`__. Included in
    this declaration are the function name, description, parameters and
    response type. This FunctionDeclaration is a representation of a
    block of code that can be used as a ``Tool`` by the model and
    executed by the client.

    Attributes:
        name (str):
            Required. The name of the function to call.
            Must start with a letter or an underscore.
            Must be a-z, A-Z, 0-9, or contain underscores,
            dots and dashes, with a maximum length of 64.
        description (str):
            Optional. Description and purpose of the
            function. Model uses it to decide how and
            whether to call the function.
        parameters (google.cloud.aiplatform_v1.types.Schema):
            Optional. Describes the parameters to this
            function in JSON Schema Object format. Reflects
            the Open API 3.03 Parameter Object. string Key:
            the name of the parameter. Parameter names are
            case sensitive. Schema Value: the Schema
            defining the type used for the parameter. For
            function with no parameters, this can be left
            unset. Parameter names must start with a letter
            or an underscore and must only contain chars
            a-z, A-Z, 0-9, or underscores with a maximum
            length of 64. Example with 1 required and 1
            optional parameter: type: OBJECT properties:

             param1:

               type: STRING
             param2:

               type: INTEGER
            required:

             - param1
        parameters_json_schema (google.protobuf.struct_pb2.Value):
            Optional. Describes the parameters to the function in JSON
            Schema format. The schema must describe an object where the
            properties are the parameters to the function. For example:

            ::

               {
                 "type": "object",
                 "properties": {
                   "name": { "type": "string" },
                   "age": { "type": "integer" }
                 },
                 "additionalProperties": false,
                 "required": ["name", "age"],
                 "propertyOrdering": ["name", "age"]
               }

            This field is mutually exclusive with ``parameters``.
        response (google.cloud.aiplatform_v1.types.Schema):
            Optional. Describes the output from this
            function in JSON Schema format. Reflects the
            Open API 3.03 Response Object. The Schema
            defines the type used for the response value of
            the function.
        response_json_schema (google.protobuf.struct_pb2.Value):
            Optional. Describes the output from this function in JSON
            Schema format. The value specified by the schema is the
            response value of the function.

            This field is mutually exclusive with ``response``.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    description: str = proto.Field(
        proto.STRING,
        number=2,
    )
    parameters: openapi.Schema = proto.Field(
        proto.MESSAGE,
        number=3,
        message=openapi.Schema,
    )
    parameters_json_schema: struct_pb2.Value = proto.Field(
        proto.MESSAGE,
        number=5,
        message=struct_pb2.Value,
    )
    response: openapi.Schema = proto.Field(
        proto.MESSAGE,
        number=4,
        message=openapi.Schema,
    )
    response_json_schema: struct_pb2.Value = proto.Field(
        proto.MESSAGE,
        number=6,
        message=struct_pb2.Value,
    )


class FunctionCall(proto.Message):
    r"""A predicted [FunctionCall] returned from the model that contains a
    string representing the [FunctionDeclaration.name] and a structured
    JSON object containing the parameters and their values.

    Attributes:
        name (str):
            Optional. The name of the function to call. Matches
            [FunctionDeclaration.name].
        args (google.protobuf.struct_pb2.Struct):
            Optional. The function parameters and values in JSON object
            format. See [FunctionDeclaration.parameters] for parameter
            details.
        partial_args (MutableSequence[google.cloud.aiplatform_v1.types.PartialArg]):
            Optional. The partial argument value of the
            function call. If provided, represents the
            arguments/fields that are streamed
            incrementally.
        will_continue (bool):
            Optional. Whether this is the last part of
            the FunctionCall. If true, another partial
            message for the current FunctionCall is expected
            to follow.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    args: struct_pb2.Struct = proto.Field(
        proto.MESSAGE,
        number=2,
        message=struct_pb2.Struct,
    )
    partial_args: MutableSequence["PartialArg"] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message="PartialArg",
    )
    will_continue: bool = proto.Field(
        proto.BOOL,
        number=5,
    )


class PartialArg(proto.Message):
    r"""Partial argument value of the function call.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        null_value (google.protobuf.struct_pb2.NullValue):
            Optional. Represents a null value.

            This field is a member of `oneof`_ ``delta``.
        number_value (float):
            Optional. Represents a double value.

            This field is a member of `oneof`_ ``delta``.
        string_value (str):
            Optional. Represents a string value.

            This field is a member of `oneof`_ ``delta``.
        bool_value (bool):
            Optional. Represents a boolean value.

            This field is a member of `oneof`_ ``delta``.
        json_path (str):
            Required. A JSON Path (RFC 9535) to the argument being
            streamed. https://datatracker.ietf.org/doc/html/rfc9535.
            e.g. "$.foo.bar[0].data".
        will_continue (bool):
            Optional. Whether this is not the last part of the same
            json_path. If true, another PartialArg message for the
            current json_path is expected to follow.
    """

    null_value: struct_pb2.NullValue = proto.Field(
        proto.ENUM,
        number=2,
        oneof="delta",
        enum=struct_pb2.NullValue,
    )
    number_value: float = proto.Field(
        proto.DOUBLE,
        number=3,
        oneof="delta",
    )
    string_value: str = proto.Field(
        proto.STRING,
        number=4,
        oneof="delta",
    )
    bool_value: bool = proto.Field(
        proto.BOOL,
        number=5,
        oneof="delta",
    )
    json_path: str = proto.Field(
        proto.STRING,
        number=1,
    )
    will_continue: bool = proto.Field(
        proto.BOOL,
        number=6,
    )


class FunctionResponsePart(proto.Message):
    r"""A datatype containing media that is part of a ``FunctionResponse``
    message.

    A ``FunctionResponsePart`` consists of data which has an associated
    datatype. A ``FunctionResponsePart`` can only contain one of the
    accepted types in ``FunctionResponsePart.data``.

    A ``FunctionResponsePart`` must have a fixed IANA MIME type
    identifying the type and subtype of the media if the ``inline_data``
    field is filled with raw bytes.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        inline_data (google.cloud.aiplatform_v1.types.FunctionResponseBlob):
            Inline media bytes.

            This field is a member of `oneof`_ ``data``.
        file_data (google.cloud.aiplatform_v1.types.FunctionResponseFileData):
            URI based data.

            This field is a member of `oneof`_ ``data``.
    """

    inline_data: "FunctionResponseBlob" = proto.Field(
        proto.MESSAGE,
        number=1,
        oneof="data",
        message="FunctionResponseBlob",
    )
    file_data: "FunctionResponseFileData" = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="data",
        message="FunctionResponseFileData",
    )


class FunctionResponseBlob(proto.Message):
    r"""Raw media bytes for function response.

    Text should not be sent as raw bytes, use the 'text' field.

    Attributes:
        mime_type (str):
            Required. The IANA standard MIME type of the
            source data.
        data (bytes):
            Required. Raw bytes.
        display_name (str):
            Optional. Display name of the blob.

            Used to provide a label or filename to distinguish blobs.

            This field is only returned in PromptMessage for prompt
            management. It is currently used in the Gemini
            GenerateContent calls only when server side tools
            (code_execution, google_search, and url_context) are
            enabled.
    """

    mime_type: str = proto.Field(
        proto.STRING,
        number=1,
    )
    data: bytes = proto.Field(
        proto.BYTES,
        number=2,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=4,
    )


class FunctionResponseFileData(proto.Message):
    r"""URI based data for function response.

    Attributes:
        mime_type (str):
            Required. The IANA standard MIME type of the
            source data.
        file_uri (str):
            Required. URI.
        display_name (str):
            Optional. Display name of the file data.

            Used to provide a label or filename to distinguish file
            datas.

            This field is only returned in PromptMessage for prompt
            management. It is currently used in the Gemini
            GenerateContent calls only when server side tools
            (code_execution, google_search, and url_context) are
            enabled.
    """

    mime_type: str = proto.Field(
        proto.STRING,
        number=1,
    )
    file_uri: str = proto.Field(
        proto.STRING,
        number=2,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=3,
    )


class FunctionResponse(proto.Message):
    r"""The result output from a [FunctionCall] that contains a string
    representing the [FunctionDeclaration.name] and a structured JSON
    object containing any output from the function is used as context to
    the model. This should contain the result of a [FunctionCall] made
    based on model prediction.

    Attributes:
        name (str):
            Required. The name of the function to call. Matches
            [FunctionDeclaration.name] and [FunctionCall.name].
        response (google.protobuf.struct_pb2.Struct):
            Required. The function response in JSON
            object format. Use "output" key to specify
            function output and "error" key to specify error
            details (if any). If "output" and "error" keys
            are not specified, then whole "response" is
            treated as function output.
        parts (MutableSequence[google.cloud.aiplatform_v1.types.FunctionResponsePart]):
            Optional. Ordered ``Parts`` that constitute a function
            response. Parts may have different IANA MIME types.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    response: struct_pb2.Struct = proto.Field(
        proto.MESSAGE,
        number=2,
        message=struct_pb2.Struct,
    )
    parts: MutableSequence["FunctionResponsePart"] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message="FunctionResponsePart",
    )


class ExecutableCode(proto.Message):
    r"""Code generated by the model that is meant to be executed, and the
    result returned to the model.

    Generated when using the [FunctionDeclaration] tool and
    [FunctionCallingConfig] mode is set to [Mode.CODE].

    Attributes:
        language (google.cloud.aiplatform_v1.types.ExecutableCode.Language):
            Required. Programming language of the ``code``.
        code (str):
            Required. The code to be executed.
    """

    class Language(proto.Enum):
        r"""Supported programming languages for the generated code.

        Values:
            LANGUAGE_UNSPECIFIED (0):
                Unspecified language. This value should not
                be used.
            PYTHON (1):
                Python >= 3.10, with numpy and simpy
                available.
        """

        LANGUAGE_UNSPECIFIED = 0
        PYTHON = 1

    language: Language = proto.Field(
        proto.ENUM,
        number=1,
        enum=Language,
    )
    code: str = proto.Field(
        proto.STRING,
        number=2,
    )


class CodeExecutionResult(proto.Message):
    r"""Result of executing the [ExecutableCode].

    Always follows a ``part`` containing the [ExecutableCode].

    Attributes:
        outcome (google.cloud.aiplatform_v1.types.CodeExecutionResult.Outcome):
            Required. Outcome of the code execution.
        output (str):
            Optional. Contains stdout when code execution
            is successful, stderr or other description
            otherwise.
    """

    class Outcome(proto.Enum):
        r"""Enumeration of possible outcomes of the code execution.

        Values:
            OUTCOME_UNSPECIFIED (0):
                Unspecified status. This value should not be
                used.
            OUTCOME_OK (1):
                Code execution completed successfully.
            OUTCOME_FAILED (2):
                Code execution finished but with a failure. ``stderr``
                should contain the reason.
            OUTCOME_DEADLINE_EXCEEDED (3):
                Code execution ran for too long, and was
                cancelled. There may or may not be a partial
                output present.
        """

        OUTCOME_UNSPECIFIED = 0
        OUTCOME_OK = 1
        OUTCOME_FAILED = 2
        OUTCOME_DEADLINE_EXCEEDED = 3

    outcome: Outcome = proto.Field(
        proto.ENUM,
        number=1,
        enum=Outcome,
    )
    output: str = proto.Field(
        proto.STRING,
        number=2,
    )


class Retrieval(proto.Message):
    r"""Defines a retrieval tool that model can call to access
    external knowledge.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        vertex_ai_search (google.cloud.aiplatform_v1.types.VertexAISearch):
            Set to use data source powered by Vertex AI
            Search.

            This field is a member of `oneof`_ ``source``.
        vertex_rag_store (google.cloud.aiplatform_v1.types.VertexRagStore):
            Set to use data source powered by Vertex RAG
            store. User data is uploaded via the
            VertexRagDataService.

            This field is a member of `oneof`_ ``source``.
        disable_attribution (bool):
            Optional. Deprecated. This option is no
            longer supported.
    """

    vertex_ai_search: "VertexAISearch" = proto.Field(
        proto.MESSAGE,
        number=2,
        oneof="source",
        message="VertexAISearch",
    )
    vertex_rag_store: "VertexRagStore" = proto.Field(
        proto.MESSAGE,
        number=4,
        oneof="source",
        message="VertexRagStore",
    )
    disable_attribution: bool = proto.Field(
        proto.BOOL,
        number=3,
    )


class VertexRagStore(proto.Message):
    r"""Retrieve from Vertex RAG Store for grounding.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        rag_resources (MutableSequence[google.cloud.aiplatform_v1.types.VertexRagStore.RagResource]):
            Optional. The representation of the rag
            source. It can be used to specify corpus only or
            ragfiles. Currently only support one corpus or
            multiple files from one corpus. In the future we
            may open up multiple corpora support.
        similarity_top_k (int):
            Optional. Number of top k results to return
            from the selected corpora.

            This field is a member of `oneof`_ ``_similarity_top_k``.
        vector_distance_threshold (float):
            Optional. Only return results with vector
            distance smaller than the threshold.

            This field is a member of `oneof`_ ``_vector_distance_threshold``.
        rag_retrieval_config (google.cloud.aiplatform_v1.types.RagRetrievalConfig):
            Optional. The retrieval config for the Rag
            query.
    """

    class RagResource(proto.Message):
        r"""The definition of the Rag resource.

        Attributes:
            rag_corpus (str):
                Optional. RagCorpora resource name. Format:
                ``projects/{project}/locations/{location}/ragCorpora/{rag_corpus}``
            rag_file_ids (MutableSequence[str]):
                Optional. rag_file_id. The files should be in the same
                rag_corpus set in rag_corpus field.
        """

        rag_corpus: str = proto.Field(
            proto.STRING,
            number=1,
        )
        rag_file_ids: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=2,
        )

    rag_resources: MutableSequence[RagResource] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message=RagResource,
    )
    similarity_top_k: int = proto.Field(
        proto.INT32,
        number=2,
        optional=True,
    )
    vector_distance_threshold: float = proto.Field(
        proto.DOUBLE,
        number=3,
        optional=True,
    )
    rag_retrieval_config: "RagRetrievalConfig" = proto.Field(
        proto.MESSAGE,
        number=6,
        message="RagRetrievalConfig",
    )


class VertexAISearch(proto.Message):
    r"""Retrieve from Vertex AI Search datastore or engine for
    grounding. datastore and engine are mutually exclusive. See
    https://cloud.google.com/products/agent-builder

    Attributes:
        datastore (str):
            Optional. Fully-qualified Vertex AI Search data store
            resource ID. Format:
            ``projects/{project}/locations/{location}/collections/{collection}/dataStores/{dataStore}``
        engine (str):
            Optional. Fully-qualified Vertex AI Search engine resource
            ID. Format:
            ``projects/{project}/locations/{location}/collections/{collection}/engines/{engine}``
        max_results (int):
            Optional. Number of search results to return
            per query. The default value is 10.
            The maximumm allowed value is 10.
        filter (str):
            Optional. Filter strings to be passed to the
            search API.
        data_store_specs (MutableSequence[google.cloud.aiplatform_v1.types.VertexAISearch.DataStoreSpec]):
            Specifications that define the specific
            DataStores to be searched, along with
            configurations for those data stores. This is
            only considered for Engines with multiple data
            stores.
            It should only be set if engine is used.
    """

    class DataStoreSpec(proto.Message):
        r"""Define data stores within engine to filter on in a search
        call and configurations for those data stores. For more
        information, see
        https://cloud.google.com/generative-ai-app-builder/docs/reference/rpc/google.cloud.discoveryengine.v1#datastorespec

        Attributes:
            data_store (str):
                Full resource name of DataStore, such as Format:
                ``projects/{project}/locations/{location}/collections/{collection}/dataStores/{dataStore}``
            filter (str):
                Optional. Filter specification to filter documents in the
                data store specified by data_store field. For more
                information on filtering, see
                `Filtering <https://cloud.google.com/generative-ai-app-builder/docs/filter-search-metadata>`__
        """

        data_store: str = proto.Field(
            proto.STRING,
            number=1,
        )
        filter: str = proto.Field(
            proto.STRING,
            number=2,
        )

    datastore: str = proto.Field(
        proto.STRING,
        number=1,
    )
    engine: str = proto.Field(
        proto.STRING,
        number=2,
    )
    max_results: int = proto.Field(
        proto.INT32,
        number=3,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=4,
    )
    data_store_specs: MutableSequence[DataStoreSpec] = proto.RepeatedField(
        proto.MESSAGE,
        number=5,
        message=DataStoreSpec,
    )


class GoogleSearchRetrieval(proto.Message):
    r"""Tool to retrieve public web data for grounding, powered by
    Google.

    Attributes:
        dynamic_retrieval_config (google.cloud.aiplatform_v1.types.DynamicRetrievalConfig):
            Specifies the dynamic retrieval configuration
            for the given source.
    """

    dynamic_retrieval_config: "DynamicRetrievalConfig" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="DynamicRetrievalConfig",
    )


class GoogleMaps(proto.Message):
    r"""Tool to retrieve public maps data for grounding, powered by
    Google.

    Attributes:
        enable_widget (bool):
            If true, include the widget context token in
            the response.
    """

    enable_widget: bool = proto.Field(
        proto.BOOL,
        number=1,
    )


class EnterpriseWebSearch(proto.Message):
    r"""Tool to search public web data, powered by Vertex AI Search
    and Sec4 compliance.


    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        exclude_domains (MutableSequence[str]):
            Optional. List of domains to be excluded from
            the search results. The default limit is 2000
            domains.
        blocking_confidence (google.cloud.aiplatform_v1.types.Tool.PhishBlockThreshold):
            Optional. Sites with confidence level chosen
            & above this value will be blocked from the
            search results.

            This field is a member of `oneof`_ ``_blocking_confidence``.
    """

    exclude_domains: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=1,
    )
    blocking_confidence: "Tool.PhishBlockThreshold" = proto.Field(
        proto.ENUM,
        number=2,
        optional=True,
        enum="Tool.PhishBlockThreshold",
    )


class DynamicRetrievalConfig(proto.Message):
    r"""Describes the options to customize dynamic retrieval.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        mode (google.cloud.aiplatform_v1.types.DynamicRetrievalConfig.Mode):
            The mode of the predictor to be used in
            dynamic retrieval.
        dynamic_threshold (float):
            Optional. The threshold to be used in dynamic
            retrieval. If not set, a system default value is
            used.

            This field is a member of `oneof`_ ``_dynamic_threshold``.
    """

    class Mode(proto.Enum):
        r"""The mode of the predictor to be used in dynamic retrieval.

        Values:
            MODE_UNSPECIFIED (0):
                Always trigger retrieval.
            MODE_DYNAMIC (1):
                Run retrieval only when system decides it is
                necessary.
        """

        MODE_UNSPECIFIED = 0
        MODE_DYNAMIC = 1

    mode: Mode = proto.Field(
        proto.ENUM,
        number=1,
        enum=Mode,
    )
    dynamic_threshold: float = proto.Field(
        proto.FLOAT,
        number=2,
        optional=True,
    )


class ToolConfig(proto.Message):
    r"""Tool config. This config is shared for all tools provided in
    the request.

    Attributes:
        function_calling_config (google.cloud.aiplatform_v1.types.FunctionCallingConfig):
            Optional. Function calling config.
        retrieval_config (google.cloud.aiplatform_v1.types.RetrievalConfig):
            Optional. Retrieval config.
    """

    function_calling_config: "FunctionCallingConfig" = proto.Field(
        proto.MESSAGE,
        number=1,
        message="FunctionCallingConfig",
    )
    retrieval_config: "RetrievalConfig" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="RetrievalConfig",
    )


class FunctionCallingConfig(proto.Message):
    r"""Function calling config.

    Attributes:
        mode (google.cloud.aiplatform_v1.types.FunctionCallingConfig.Mode):
            Optional. Function calling mode.
        allowed_function_names (MutableSequence[str]):
            Optional. Function names to call. Only set when the Mode is
            ANY. Function names should match [FunctionDeclaration.name].
            With mode set to ANY, model will predict a function call
            from the set of function names provided.
        stream_function_call_arguments (bool):
            Optional. When set to true, arguments of a single function
            call will be streamed out in multiple
            parts/contents/responses. Partial parameter results will be
            returned in the [FunctionCall.partial_args] field.
    """

    class Mode(proto.Enum):
        r"""Function calling mode.

        Values:
            MODE_UNSPECIFIED (0):
                Unspecified function calling mode. This value
                should not be used.
            AUTO (1):
                Default model behavior, model decides to
                predict either function calls or natural
                language response.
            ANY (2):
                Model is constrained to always predicting function calls
                only. If "allowed_function_names" are set, the predicted
                function calls will be limited to any one of
                "allowed_function_names", else the predicted function calls
                will be any one of the provided "function_declarations".
            NONE (3):
                Model will not predict any function calls.
                Model behavior is same as when not passing any
                function declarations.
        """

        MODE_UNSPECIFIED = 0
        AUTO = 1
        ANY = 2
        NONE = 3

    mode: Mode = proto.Field(
        proto.ENUM,
        number=1,
        enum=Mode,
    )
    allowed_function_names: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=2,
    )
    stream_function_call_arguments: bool = proto.Field(
        proto.BOOL,
        number=4,
    )


class RetrievalConfig(proto.Message):
    r"""Retrieval config.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        lat_lng (google.type.latlng_pb2.LatLng):
            The location of the user.

            This field is a member of `oneof`_ ``_lat_lng``.
        language_code (str):
            The language code of the user.

            This field is a member of `oneof`_ ``_language_code``.
    """

    lat_lng: latlng_pb2.LatLng = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message=latlng_pb2.LatLng,
    )
    language_code: str = proto.Field(
        proto.STRING,
        number=2,
        optional=True,
    )


class RagRetrievalConfig(proto.Message):
    r"""Specifies the context retrieval config.

    Attributes:
        top_k (int):
            Optional. The number of contexts to retrieve.
        filter (google.cloud.aiplatform_v1.types.RagRetrievalConfig.Filter):
            Optional. Config for filters.
        ranking (google.cloud.aiplatform_v1.types.RagRetrievalConfig.Ranking):
            Optional. Config for ranking and reranking.
    """

    class Filter(proto.Message):
        r"""Config for filters.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            vector_distance_threshold (float):
                Optional. Only returns contexts with vector
                distance smaller than the threshold.

                This field is a member of `oneof`_ ``vector_db_threshold``.
            vector_similarity_threshold (float):
                Optional. Only returns contexts with vector
                similarity larger than the threshold.

                This field is a member of `oneof`_ ``vector_db_threshold``.
            metadata_filter (str):
                Optional. String for metadata filtering.
        """

        vector_distance_threshold: float = proto.Field(
            proto.DOUBLE,
            number=3,
            oneof="vector_db_threshold",
        )
        vector_similarity_threshold: float = proto.Field(
            proto.DOUBLE,
            number=4,
            oneof="vector_db_threshold",
        )
        metadata_filter: str = proto.Field(
            proto.STRING,
            number=2,
        )

    class Ranking(proto.Message):
        r"""Config for ranking and reranking.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            rank_service (google.cloud.aiplatform_v1.types.RagRetrievalConfig.Ranking.RankService):
                Optional. Config for Rank Service.

                This field is a member of `oneof`_ ``ranking_config``.
            llm_ranker (google.cloud.aiplatform_v1.types.RagRetrievalConfig.Ranking.LlmRanker):
                Optional. Config for LlmRanker.

                This field is a member of `oneof`_ ``ranking_config``.
        """

        class RankService(proto.Message):
            r"""Config for Rank Service.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                model_name (str):
                    Optional. The model name of the rank service. Format:
                    ``semantic-ranker-512@latest``

                    This field is a member of `oneof`_ ``_model_name``.
            """

            model_name: str = proto.Field(
                proto.STRING,
                number=1,
                optional=True,
            )

        class LlmRanker(proto.Message):
            r"""Config for LlmRanker.

            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                model_name (str):
                    Optional. The model name used for ranking. Format:
                    ``gemini-1.5-pro``

                    This field is a member of `oneof`_ ``_model_name``.
            """

            model_name: str = proto.Field(
                proto.STRING,
                number=1,
                optional=True,
            )

        rank_service: "RagRetrievalConfig.Ranking.RankService" = proto.Field(
            proto.MESSAGE,
            number=1,
            oneof="ranking_config",
            message="RagRetrievalConfig.Ranking.RankService",
        )
        llm_ranker: "RagRetrievalConfig.Ranking.LlmRanker" = proto.Field(
            proto.MESSAGE,
            number=3,
            oneof="ranking_config",
            message="RagRetrievalConfig.Ranking.LlmRanker",
        )

    top_k: int = proto.Field(
        proto.INT32,
        number=1,
    )
    filter: Filter = proto.Field(
        proto.MESSAGE,
        number=3,
        message=Filter,
    )
    ranking: Ranking = proto.Field(
        proto.MESSAGE,
        number=4,
        message=Ranking,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2015 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define API Jobs."""

from google.cloud.bigquery.job.base import _AsyncJob
from google.cloud.bigquery.job.base import _error_result_to_exception
from google.cloud.bigquery.job.base import _DONE_STATE
from google.cloud.bigquery.job.base import _JobConfig
from google.cloud.bigquery.job.base import _JobReference
from google.cloud.bigquery.job.base import ReservationUsage
from google.cloud.bigquery.job.base import ScriptStatistics
from google.cloud.bigquery.job.base import ScriptStackFrame
from google.cloud.bigquery.job.base import TransactionInfo
from google.cloud.bigquery.job.base import UnknownJob
from google.cloud.bigquery.job.copy_ import CopyJob
from google.cloud.bigquery.job.copy_ import CopyJobConfig
from google.cloud.bigquery.job.copy_ import OperationType
from google.cloud.bigquery.job.extract import ExtractJob
from google.cloud.bigquery.job.extract import ExtractJobConfig
from google.cloud.bigquery.job.load import LoadJob
from google.cloud.bigquery.job.load import LoadJobConfig
from google.cloud.bigquery.job.query import _contains_order_by
from google.cloud.bigquery.job.query import DmlStats
from google.cloud.bigquery.job.query import QueryJob
from google.cloud.bigquery.job.query import QueryJobConfig
from google.cloud.bigquery.job.query import QueryPlanEntry
from google.cloud.bigquery.job.query import QueryPlanEntryStep
from google.cloud.bigquery.job.query import ScriptOptions
from google.cloud.bigquery.job.query import TimelineEntry
from google.cloud.bigquery.job.query import IncrementalResultStats
from google.cloud.bigquery.enums import Compression
from google.cloud.bigquery.enums import CreateDisposition
from google.cloud.bigquery.enums import DestinationFormat
from google.cloud.bigquery.enums import Encoding
from google.cloud.bigquery.enums import QueryPriority
from google.cloud.bigquery.enums import SchemaUpdateOption
from google.cloud.bigquery.enums import SourceFormat
from google.cloud.bigquery.enums import WriteDisposition


# Include classes previously in job.py for backwards compatibility.
__all__ = [
    "_AsyncJob",
    "_error_result_to_exception",
    "_DONE_STATE",
    "_JobConfig",
    "_JobReference",
    "ReservationUsage",
    "ScriptStatistics",
    "ScriptStackFrame",
    "UnknownJob",
    "CopyJob",
    "CopyJobConfig",
    "OperationType",
    "ExtractJob",
    "ExtractJobConfig",
    "LoadJob",
    "LoadJobConfig",
    "_contains_order_by",
    "DmlStats",
    "QueryJob",
    "QueryJobConfig",
    "QueryPlanEntry",
    "QueryPlanEntryStep",
    "ScriptOptions",
    "TimelineEntry",
    "Compression",
    "CreateDisposition",
    "DestinationFormat",
    "Encoding",
    "QueryPriority",
    "SchemaUpdateOption",
    "SourceFormat",
    "TransactionInfo",
    "WriteDisposition",
    "IncrementalResultStats",
]



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1beta1.types import value as gca_value
from google.protobuf import struct_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "ArtifactTypeSchema",
        "RuntimeArtifact",
    },
)


class ArtifactTypeSchema(proto.Message):
    r"""The definition of a artifact type in MLMD.

    This message has `oneof`_ fields (mutually exclusive fields).
    For each oneof, at most one member field can be set at the same time.
    Setting any member of the oneof automatically clears all other
    members.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        schema_title (str):
            The name of the type. The format of the title must be:
            ``<namespace>.<title>``. Examples:

            - ``aiplatform.Model``
            - ``acme.CustomModel`` When this field is set, the type must
              be pre-registered in the MLMD store.

            This field is a member of `oneof`_ ``kind``.
        schema_uri (str):
            Points to a YAML file stored on Cloud Storage describing the
            format. Deprecated. Use
            [PipelineArtifactTypeSchema.schema_title][] or
            [PipelineArtifactTypeSchema.instance_schema][] instead.

            This field is a member of `oneof`_ ``kind``.
        instance_schema (str):
            Contains a raw YAML string, describing the
            format of the properties of the type.

            This field is a member of `oneof`_ ``kind``.
        schema_version (str):
            The schema version of the artifact. If the
            value is not set, it defaults to the latest
            version in the system.
    """

    schema_title: str = proto.Field(
        proto.STRING,
        number=1,
        oneof="kind",
    )
    schema_uri: str = proto.Field(
        proto.STRING,
        number=2,
        oneof="kind",
    )
    instance_schema: str = proto.Field(
        proto.STRING,
        number=3,
        oneof="kind",
    )
    schema_version: str = proto.Field(
        proto.STRING,
        number=4,
    )


class RuntimeArtifact(proto.Message):
    r"""The definition of a runtime artifact.

    Attributes:
        name (str):
            The name of an artifact.
        type_ (google.cloud.aiplatform_v1beta1.types.ArtifactTypeSchema):
            The type of the artifact.
        uri (str):
            The URI of the artifact.
        properties (MutableMapping[str, google.cloud.aiplatform_v1beta1.types.Value]):
            The properties of the artifact. Deprecated. Use
            [RuntimeArtifact.metadata][google.cloud.aiplatform.v1beta1.RuntimeArtifact.metadata]
            instead.
        custom_properties (MutableMapping[str, google.cloud.aiplatform_v1beta1.types.Value]):
            The custom properties of the artifact. Deprecated. Use
            [RuntimeArtifact.metadata][google.cloud.aiplatform.v1beta1.RuntimeArtifact.metadata]
            instead.
        metadata (google.protobuf.struct_pb2.Struct):
            Properties of the Artifact.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    type_: "ArtifactTypeSchema" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="ArtifactTypeSchema",
    )
    uri: str = proto.Field(
        proto.STRING,
        number=3,
    )
    properties: MutableMapping[str, gca_value.Value] = proto.MapField(
        proto.STRING,
        proto.MESSAGE,
        number=4,
        message=gca_value.Value,
    )
    custom_properties: MutableMapping[str, gca_value.Value] = proto.MapField(
        proto.STRING,
        proto.MESSAGE,
        number=5,
        message=gca_value.Value,
    )
    metadata: struct_pb2.Struct = proto.Field(
        proto.MESSAGE,
        number=6,
        message=struct_pb2.Struct,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



import asyncio
import logging
import os
import httpx
import bmasterai
from fastmcp import FastMCP, Context

# Initialize BMasterAI Logging
bmasterai.configure_logging(log_file='weather_mcp.log', json_log_file='weather_mcp.jsonl')
logger = bmasterai.get_logger()

# Configure logging - REMOVED to use BMasterAI
# logging.basicConfig(format="[%(levelname)s]: %(message)s", level=logging.INFO)
# logger = logging.getLogger("weather-mcp")

# Initialize FastMCP server
mcp = FastMCP("Weather MCP Server 🌤️")

@mcp.tool()
async def get_weather_forecast(city: str, days: int = 3) -> dict:
    """
    Get weather forecast for a specific city.

    Args:
        city: The name of the city (e.g., "Indianapolis", "San Francisco").
        days: Number of forecast days (default: 3).

    Returns:
        A dictionary containing weather forecast data.
    """
    logger.logger.info(f"--- 🌤️ Tool: get_weather_forecast called for {city} ({days} days) ---")
    logger.log_event("weather-mcp", bmasterai.EventType.TOOL_USE, f"Fetching weather for {city}", metadata={"city": city, "days": days})

    try:
        # 1. Geocoding
        geo_url = "https://geocoding-api.open-meteo.com/v1/search"
        params = {"name": city, "count": 1, "language": "en", "format": "json"}
        
        async with httpx.AsyncClient() as client:
            resp = await client.get(geo_url, params=params)
            data = resp.json()
            
            if not data.get("results"):
                 return {"error": f"City '{city}' not found."}
                 
            lat = data["results"][0]["latitude"]
            lon = data["results"][0]["longitude"]
            city_name = data["results"][0]["name"]
            country = data["results"][0].get("country", "Unknown")
            
            logger.logger.info(f"Resolved {city} to {lat}, {lon}")

            # Now fetch weather forecast
            weather_response = await client.get(
                "https://api.open-meteo.com/v1/forecast",
                params={
                    "latitude": lat,
                    "longitude": lon,
                    "daily": "weather_code,temperature_2m_max,temperature_2m_min",
                    "forecast_days": days,
                    "timezone": "auto",
                },
            )
            weather_response.raise_for_status()
            weather_data = weather_response.json()
            
            daily = weather_data.get("daily", {})
            units = weather_data.get("daily_units", {})
            
            # Helper to convert C to F
            def c_to_f(c_list):
                return [round((t * 9/5) + 32, 1) if t is not None else None for t in c_list]

            # Add F data
            if "temperature_2m_max" in daily:
                daily["temperature_2m_max_f"] = c_to_f(daily["temperature_2m_max"])
                units["temperature_2m_max_f"] = "°F"
                
            if "temperature_2m_min" in daily:
                daily["temperature_2m_min_f"] = c_to_f(daily["temperature_2m_min"])
                units["temperature_2m_min_f"] = "°F"

            return {
                "location": f"{city_name}, {country}",
                "forecast": daily,
                "units": units
            }

    except httpx.HTTPError as e:
        logger.logger.error(f"❌ API request failed: {e}")
        return {"error": f"Failed to fetch weather data: {str(e)}"}
    except Exception as e:
        logger.logger.error(f"❌ Unexpected error: {e}")
        return {"error": f"An unexpected error occurred: {str(e)}"}

if __name__ == "__main__":
    logger.logger.info(f"🚀 Weather MCP server started on port {os.getenv('PORT', 8080)}")
    asyncio.run(
        mcp.run_async(
            transport="http",
            host="0.0.0.0",
            port=int(os.getenv("PORT", 8080)),
        )
    )
